from __future__ import annotations
import asyncio
import base64
import io
import json
import logging
import os
import platform
import re
import textwrap
import uuid
from io import BytesIO
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Type, TypeVar
from dotenv import load_dotenv
from google.api_core.exceptions import ResourceExhausted
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.messages import AIMessage, BaseMessage, HumanMessage, SystemMessage
from lmnr import observe
from openai import RateLimitError
from PIL import Image, ImageDraw, ImageFont
from pydantic import BaseModel, ValidationError
from browser_use.agent.message_manager.service import MessageManager
from browser_use.agent.prompts import AgentMessagePrompt, PlannerPrompt, SystemPrompt
from browser_use.agent.views import ActionResult, AgentError, AgentHistory, AgentHistoryList, AgentOutput, AgentStepInfo
from browser_use.browser.browser import Browser
from browser_use.browser.context import BrowserContext
from browser_use.browser.views import BrowserState, BrowserStateHistory
from browser_use.controller.registry.views import ActionModel
from browser_use.controller.service import Controller
from browser_use.dom.history_tree_processor.service import DOMHistoryElement, HistoryTreeProcessor
from browser_use.telemetry.service import ProductTelemetry
from browser_use.telemetry.views import AgentEndTelemetryEvent, AgentRunTelemetryEvent, AgentStepTelemetryEvent
from browser_use.utils import time_execution_async

load_dotenv()
logger = logging.getLogger(__name__)
T = TypeVar('T', bound=BaseModel)

class Agent:
	def __init__(
		self,
		task: str,
		llm: BaseChatModel,
		browser: Browser | None = None,
		browser_context: BrowserContext | None = None,
		controller: Controller = Controller(),
		use_vision: bool = True,
		use_vision_for_planner: bool = False,
		save_conversation_path: Optional[str] = None,
		save_conversation_path_encoding: Optional[str] = 'utf-8',
		max_failures: int = 3,
		retry_delay: int = 10,
		system_prompt_class: Type[SystemPrompt] = SystemPrompt,
		max_input_tokens: int = 128000,
		validate_output: bool = False,
		message_context: Optional[str] = None,
		generate_gif: bool | str = False,
		sensitive_data: Optional[Dict[str, str]] = None,
		include_attributes: list[str] = [
			'title', 'type', 'name', 'role', 'tabindex', 'aria-label', 
			'placeholder', 'value', 'alt', 'aria-expanded'
		],
		max_error_length: int = 400,
		max_actions_per_step: int = 10,
		tool_call_in_content: bool = True,
		initial_actions: Optional[List[Dict[str, Dict[str, Any]]]] = None,
		register_new_step_callback: Callable[['BrowserState', 'AgentOutput', int], None] | None = None,
		register_done_callback: Callable[['AgentHistoryList'], None] | None = None,
		tool_calling_method: Optional[str] = 'auto',
		page_extraction_llm: Optional[BaseChatModel] = None,
		planner_llm: Optional[BaseChatModel] = None,
		planner_interval: int = 1
	):
		self.agent_id = str(uuid.uuid4())
		self.sensitive_data = sensitive_data
		
		if not page_extraction_llm:
			self.page_extraction_llm = llm
		else:
			self.page_extraction_llm = page_extraction_llm
			
		self.task = task
		self.use_vision = use_vision
		self.use_vision_for_planner = use_vision_for_planner
		self.llm = llm
		self.save_conversation_path = save_conversation_path
		self.save_conversation_path_encoding = save_conversation_path_encoding
		self._last_result = None
		self.include_attributes = include_attributes
		self.max_error_length = max_error_length
		self.generate_gif = generate_gif
		self.planner_llm = planner_llm
		self.planning_interval = planner_interval
		self.last_plan = None
		self.controller = controller
		self.max_actions_per_step = max_actions_per_step
		self.injected_browser = browser is not None
		self.injected_browser_context = browser_context is not None
		self.message_context = message_context
		
		# Setup browser and browser context
		if browser is not None:
			self.browser = browser
		elif browser_context:
			self.browser = None
		else:
			self.browser = Browser()
			
		if browser_context:
			self.browser_context = browser_context
		elif self.browser:
			self.browser_context = BrowserContext(browser=self.browser, config=self.browser.config.new_context_config)
		else:
			self.browser = Browser()
			self.browser_context = BrowserContext(browser=self.browser)
		
		self.system_prompt_class = system_prompt_class
		self.telemetry = ProductTelemetry()
		
		# Setup actions and models
		self._setup_action_models()
		self._set_version_and_source()
		self.max_input_tokens = max_input_tokens
		self._set_model_names()
		
		self.tool_calling_method = self.set_tool_calling_method(tool_calling_method)
		
		# Initialize message manager
		self.message_manager = MessageManager(
			llm=self.llm,
			task=self.task,
			action_descriptions=self.controller.registry.get_prompt_description(),
			system_prompt_class=self.system_prompt_class,
			max_input_tokens=self.max_input_tokens,
			include_attributes=self.include_attributes,
			max_error_length=self.max_error_length,
			max_actions_per_step=self.max_actions_per_step,
			message_context=self.message_context,
			sensitive_data=self.sensitive_data
		)
		
		self.register_new_step_callback = register_new_step_callback
		self.register_done_callback = register_done_callback
		self.history = AgentHistoryList(history=[])
		self.n_steps = 1
		self.consecutive_failures = 0
		self.max_failures = max_failures
		self.retry_delay = retry_delay
		self.validate_output = validate_output
		
		if initial_actions:
			self.initial_actions = self._convert_initial_actions(initial_actions)
		else:
			self.initial_actions = None
			
		if save_conversation_path:
			logger.info(f'Saving conversation to {save_conversation_path}')
			
		self._paused = False
		self._stopped = False
		self.action_descriptions = self.controller.registry.get_prompt_description()

	def _set_version_and_source(self):
		try:
			import pkg_resources
			version = pkg_resources.get_distribution('browser-use').version
			source = 'pip'
		except Exception:
			try:
				import subprocess
				version = subprocess.check_output(['git', 'describe', '--tags']).decode('utf-8').strip()
				source = 'git'
			except Exception:
				version = 'unknown'
				source = 'unknown'
				
		logger.debug(f'Version: {version}, Source: {source}')
		self.version = version
		self.source = source
		
	def _set_model_names(self):
		self.chat_model_library = self.llm.__class__.__name__
		
		if hasattr(self.llm, 'model_name'):
			self.model_name = self.llm.model_name
		elif hasattr(self.llm, 'model'):
			self.model_name = self.llm.model
		else:
			self.model_name = 'Unknown'
			
		if self.planner_llm:
			if hasattr(self.planner_llm, 'model_name'):
				self.planner_model_name = self.planner_llm.model_name
			elif hasattr(self.planner_llm, 'model'):
				self.planner_model_name = self.planner_llm.model
			else:
				self.planner_model_name = 'Unknown'
		else:
			self.planner_model_name = None
			
	def _setup_action_models(self):
		"""Setup dynamic action models from controller's registry"""
		self.ActionModel = self.controller.registry.create_action_model()
		self.AgentOutput = AgentOutput.type_with_custom_actions(self.ActionModel)
		
	def set_tool_calling_method(self, tool_calling_method):
		if tool_calling_method == 'auto':
			if self.chat_model_library == 'ChatGoogleGenerativeAI':
				return None
			elif self.chat_model_library == 'ChatOpenAI':
				return 'function_calling'
			elif self.chat_model_library == 'AzureChatOpenAI':
				return 'function_calling'
			else:
				return None
		else:
			return tool_calling_method
			
	def add_new_task(self, new_task):
		self.message_manager.add_new_task(new_task)
		
	def _check_if_stopped_or_paused(self):
		if self._stopped or self._paused:
			logger.debug('Agent paused after getting state')
			raise InterruptedError
		return False
		
	@observe(name='agent.step', ignore_output=True, ignore_input=True)
	@time_execution_async('--step')
	async def step(self, step_info: Optional[AgentStepInfo] = None):
		"""Execute one step of the task"""
		logger.info(f'📍 Step {self.n_steps}')
		
		state = None
		model_output = None
		result = []
		
		try:
			# Get browser state
			state = await self.browser_context.get_state()
			self._check_if_stopped_or_paused()
			
			# Add state to message history
			self.message_manager.add_state_message(state, self._last_result, step_info, self.use_vision)
			
			# Run planner at specified intervals if configured
			if self.planner_llm and self.n_steps % self.planning_interval == 0:
				plan = await self._run_planner()
				self.message_manager.add_plan(plan, position=-1)
				
			# Get all messages for model
			messages = self.message_manager.get_messages()
			self._check_if_stopped_or_paused()
			
			# Get next action from LLM
			model_output = await self.get_next_action(messages)
			
			# Callback if registered
			if self.register_new_step_callback:
				self.register_new_step_callback(state, model_output, self.n_steps)
				
			# Save conversation if path specified
			self._save_conversation(messages, model_output)
			
			# Remove state from history to avoid token bloat
			self.message_manager._remove_last_state_message()
			self._check_if_stopped_or_paused()
			
			# Add model output to message history
			self.message_manager.add_model_output(model_output)
			
		except Exception as e:
			# If message processing failed, remove state message
			self.message_manager._remove_last_state_message()
			raise e
			
		# Execute actions
		result = await self.controller.multi_act(
			model_output.action, 
			self.browser_context,
			self.page_extraction_llm,
			self.sensitive_data,
			check_break_if_paused=lambda: self._check_if_stopped_or_paused()
		)
		
		self._last_result = result
		
		# Check if task is done
		if len(result) > 0 and result[-1].is_done:
			logger.info(f'📄 Result: {result[-1].extracted_content}')
			
		# Reset failure counter on success
		self.consecutive_failures = 0
		
	async def _handle_step_error(self, error: Exception) -> list[ActionResult]:
		"""Handle all types of errors that can occur during a step"""
		include_trace = logger.isEnabledFor(logging.DEBUG)
		error_msg = AgentError.format_error(error, include_trace=include_trace)
		prefix = f'❌ Result failed {self.consecutive_failures + 1}/{self.max_failures} times:\n '
		
		if isinstance(error, (ValidationError, ValueError)):
			logger.error(f'{prefix}{error_msg}')
			
			if 'Max token limit reached' in error_msg:
				# Cut tokens from history
				self.message_manager.max_input_tokens = self.max_input_tokens - 500
				logger.info(f'Cutting tokens from history - new max input tokens: {self.message_manager.max_input_tokens}')
				self.message_manager.cut_messages()
			elif 'Could not parse response' in error_msg:
				# Add hint for proper response format
				error_msg += '\n\nReturn a valid JSON object with the required fields.'
				
			self.consecutive_failures += 1
			
		elif isinstance(error, RateLimitError) or isinstance(error, ResourceExhausted):
			logger.warning(f'{prefix}{error_msg}')
			await asyncio.sleep(self.retry_delay)
			self.consecutive_failures += 1
			
		else:
			logger.error(f'{prefix}{error_msg}')
			self.consecutive_failures += 1
			
		return [ActionResult(error=error_msg, include_in_memory=True)]
		
	def _make_history_item(self, model_output: AgentOutput | None, state: BrowserState, result: list[ActionResult]):
		"""Create and store history item"""
		interacted_elements = None
		num_results = len(result)
		
		if model_output:
			interacted_elements = AgentHistory.get_interacted_element(model_output, state.selector_map)
		else:
			interacted_elements = [None]
			
		browser_state_history = BrowserStateHistory(
			url=state.url, 
			title=state.title, 
			tabs=state.tabs, 
			interacted_element=interacted_elements, 
			screenshot=''
		)
		
		history_item = AgentHistory(
			model_output=model_output, 
			result=result, 
			state=browser_state_history
		)
		
		self.history.history.append(history_item)
		
	THINK_TAGS_REGEX = re.compile('<think>.*?</think>', re.DOTALL)
	
	def _remove_think_tags(self, text: str) -> str:
		"""Remove think tags from text"""
		return re.sub(self.THINK_TAGS_REGEX, '', text)
		
	def _convert_input_messages(self, input_messages: list[BaseMessage], model_name: Optional[str]) -> list[BaseMessage]:
		"""Convert input messages to a format that is compatible with the planner model"""
		if model_name is None:
			return input_messages
			
		if model_name == 'deepseek-reasoner' or model_name.startswith('deepseek-r1'):
			converted_messages = self.message_manager.convert_messages_for_non_function_calling_models(input_messages)
			merged_messages = self.message_manager.merge_successive_messages(converted_messages, HumanMessage)
			merged_messages = self.message_manager.merge_successive_messages(merged_messages, AIMessage)
			return merged_messages
			
		return input_messages
		
	@time_execution_async('--get_next_action')
	async def get_next_action(self, input_messages: list[BaseMessage]) -> AgentOutput:
		"""Get next action from LLM based on current state"""
		# Convert input messages to the appropriate format
		input_messages = self._convert_input_messages(input_messages, self.model_name)
		
		# Handle raw text completion models
		if self.model_name == 'deepseek-reasoner' or self.model_name.startswith('deepseek-r1'):
			model_response = self.llm.invoke(input_messages)
			model_response.content = self._remove_think_tags(model_response.content)
			
			try:
				extracted_json = self.message_manager.extract_json_from_model_output(model_response.content)
				parsed_output = self.AgentOutput(**extracted_json)
			except (ValueError, ValidationError) as e:
				logger.warning(f'Failed to parse model output: {model_response} {str(e)}')
				raise ValueError('Could not parse response.')
				
		# Handle structured output with function calling
		elif self.tool_calling_method is not None:
			structured_llm = self.llm.with_structured_output(self.AgentOutput, include_raw=True)
			response = await structured_llm.ainvoke(input_messages)
			parsed_output = response['parsed']
			
		else:
			structured_llm = self.llm.with_structured_output(self.AgentOutput, include_raw=True, method=self.tool_calling_method)
			response = await structured_llm.ainvoke(input_messages)
			parsed_output = response['parsed']
			
		if parsed_output is None:
			raise ValueError('Could not parse response.')
			
		# Limit the number of actions
		parsed_output.action = parsed_output.action[:self.max_actions_per_step]
		
		self._log_response(parsed_output)
		self.n_steps += 1
		
		return parsed_output
		
	def _log_response(self, response: AgentOutput) -> None:
		"""Log the model's response"""
		if 'Success' in response.current_state.evaluation_previous_goal:
			emoji = '👍'
		elif 'Failed' in response.current_state.evaluation_previous_goal:
			emoji = '⚠'
		else:
			emoji = '🤷'
			
		logger.debug(f'🤖 {emoji} Page summary: {response.current_state.page_summary}')
		logger.info(f'{emoji} Eval: {response.current_state.evaluation_previous_goal}')
		logger.info(f'🧠 Memory: {response.current_state.memory}')
		logger.info(f'🎯 Next goal: {response.current_state.next_goal}')
		
		for i, action in enumerate(response.action):
			logger.info(f'🛠️  Action {i + 1}/{len(response.action)}: {action.model_dump_json(exclude_unset=True)}')
		
	def _save_conversation(self, input_messages: list[BaseMessage], response: Any) -> None:
		"""Save conversation history to file if path is specified"""
		if not self.save_conversation_path:
			return None
			
		os.makedirs(os.path.dirname(self.save_conversation_path), exist_ok=True)
		
		with open(self.save_conversation_path + f'_{self.n_steps}.txt', 'w', encoding=self.save_conversation_path_encoding) as file:
			self._write_messages_to_file(file, input_messages)
			self._write_response_to_file(file, response)
			
	def _write_messages_to_file(self, file, messages: list[BaseMessage]) -> None:
		"""Write messages to conversation file"""
		for message in messages:
			file.write(f' {message.__class__.__name__} \n')
			
			if isinstance(message.content, list):
				for content_item in message.content:
					if isinstance(content_item, dict) and content_item.get('type') == 'text':
						file.write(content_item['text'].strip() + '\n')
			elif isinstance(message.content, str):
				try:
					parsed_content = json.loads(message.content)
					file.write(json.dumps(parsed_content, indent=2) + '\n')
				except json.JSONDecodeError:
					file.write(message.content.strip() + '\n')
					
			file.write('\n')
			
	def _write_response_to_file(self, file, response: Any) -> None:
		"""Write model response to conversation file"""
		file.write(' RESPONSE\n')
		file.write(json.dumps(json.loads(response.model_dump_json(exclude_unset=True)), indent=2))
		
	def _log_agent_run(self) -> None:
		"""Log the agent run"""
		logger.info(f'🚀 Starting task: {self.task}')
		logger.debug(f'Version: {self.version}, Source: {self.source}')
		
		self.telemetry.capture(
			AgentRunTelemetryEvent(
				agent_id=self.agent_id,
				use_vision=self.use_vision,
				task=self.task,
				model_name=self.model_name,
				chat_model_library=self.chat_model_library,
				version=self.version,
				source=self.source
			)
		)
		
	@observe(name='agent.run', ignore_output=True)
	@time_execution_async('--run')
	async def run(self, max_steps: int = 100) -> AgentHistoryList:
		"""Execute the task with maximum number of steps"""
		try:
			self._log_agent_run()
			
			# Execute initial actions if provided
			if self.initial_actions:
				result = await self.controller.multi_act(
					self.initial_actions,
					self.browser_context,
					False,
					self.page_extraction_llm,
					check_break_if_paused=lambda: self._check_if_stopped_or_paused()
				)
				self._last_result = result
				
			# Main execution loop
			for step in range(max_steps):
				# Check for too many failures
				if self._too_many_failures():
					break
					
				# Check control flags
				if not await self._handle_control_flags():
					break
					
				# Execute step
				await self.step()
				
				# Check if task is completed
				if self.history.is_done():
					# Validate output if enabled
					if self.validate_output and step < max_steps - 1:
						if not await self._validate_output():
							continue
							
					logger.info('✅ Task completed successfully')
					
					# Call done callback if registered
					if self.register_done_callback:
						self.register_done_callback(self.history)
						
					break
			else:
				logger.info('❌ Failed to complete task in maximum steps')
				
			return self.history
			
		except Exception:
			# Ensure proper cleanup on exception
			actions = [action.model_dump(exclude_unset=True) for action in model_output.action] if model_output else []
			
			self.telemetry.capture(
				AgentStepTelemetryEvent(
					agent_id=self.agent_id,
					step=self.n_steps,
					actions=actions,
					consecutive_failures=self.consecutive_failures,
					step_error=[r.error for r in result if r.error] if result else ['No result']
				)
			)
			
			if state:
				self._make_history_item(model_output, state, result)
				
			raise
			
		finally:
			# Capture telemetry
			actions = [action.model_dump(exclude_unset=True) for action in model_output.action] if model_output else []
			
			self.telemetry.capture(
				AgentEndTelemetryEvent(
					agent_id=self.agent_id,
					step=self.n_steps,
					actions=actions,
					consecutive_failures=self.consecutive_failures,
					step_error=[r.error for r in result if r.error] if result else ['No result']
				)
			)
			
			if state:
				self._make_history_item(model_output, state, result)
				
			# Close resources
			if not self.injected_browser_context:
				await self.browser_context.close()
				
			if not self.injected_browser and self.browser:
				await self.browser.close()
				
			# Generate GIF if enabled
			if self.generate_gif:
				output_path = 'agent_history.gif'
				if isinstance(self.generate_gif, str):
					output_path = self.generate_gif
					
				self.create_history_gif(output_path=output_path)
				
	def _too_many_failures(self) -> bool:
		"""Check if we should stop due to too many failures"""
		if self.consecutive_failures >= self.max_failures:
			logger.error(f'❌ Stopping due to {self.max_failures} consecutive failures')
			return True
		return False
		
	async def _handle_control_flags(self) -> bool:
		"""Handle pause and stop flags. Returns True if execution should continue."""
		if self._stopped:
			logger.debug('Agent stopped')
			return False
			
		while self._paused:
			await asyncio.sleep(0.2)  # Small delay to prevent CPU spinning
			if self._stopped:  # Allow stopping while paused
				return False
				
		return True
		
	async def _validate_output(self) -> bool:
		"""Validate the output of the last action is what the user wanted"""
		system_msg = (
			'You are a validator of an agent who interacts with a browser. '
			'Validate if the output of last action is what the user wanted and if the task is completed. '
			'If the task is unclear defined, you can let it pass. But if something is missing or the image does not show what was requested dont let it pass. '
			'Try to understand the page and help the model with suggestions like scroll, do x, ... to get the solution right. '
			f'Task to validate: {self.task}. Return a JSON object with 2 keys: is_valid and reason. '
			'is_valid is a boolean that indicates if the output is correct. '
			'reason is a string that explains why it is valid or not. '
			'example: {"is_valid": false, "reason": "The user wanted to search for "cat photos", but the agent searched for "dog photos" instead."}'
		)
		
		if not self.browser_context.session:
			return True
			
		state = await self.browser_context.get_state()
		prompt = AgentMessagePrompt(
			state=state,
			result=self._last_result,
			include_attributes=self.include_attributes,
			max_error_length=self.max_error_length
		)
		
		messages = [SystemMessage(content=system_msg), prompt.get_user_message(self.use_vision)]
		
		class ValidationResult(BaseModel):
			is_valid: bool
			reason: str
			
		validator = self.llm.with_structured_output(ValidationResult, include_raw=True)
		response = await validator.ainvoke(messages)
		parsed_result = response['parsed']
		
		is_valid = parsed_result.is_valid
		if not is_valid:
			logger.info(f'❌ Validator decision: {parsed_result.reason}')
			message = f'The output is not yet correct. {parsed_result.reason}.'
			self._last_result = [ActionResult(extracted_content=message, include_in_memory=True)]
		else:
			logger.info(f'✅ Validator decision: {parsed_result.reason}')
			
		return is_valid
		
	async def rerun_history(
		self,
		history: AgentHistoryList,
		max_retries: int = 3,
		skip_failures: bool = True,
		delay_between_actions: float = 2.0
	) -> list[ActionResult]:
		"""
		Rerun a saved history of actions with error handling and retry logic.

		Args:
			history: The history to replay
			max_retries: Maximum number of retries per action
			skip_failures: Whether to skip failed actions or stop execution
			delay_between_actions: Delay between actions in seconds

		Returns:
			List of action results
		"""
		if self.initial_actions:
			result = await self.controller.multi_act(
				self.initial_actions,
				self.browser_context,
				False,
				self.page_extraction_llm,
				check_break_if_paused=lambda: self._check_if_stopped_or_paused()
			)
			self._last_result = result
			
		results = []
		
		for i, history_item in enumerate(history.history):
			goal = history_item.model_output.current_state.next_goal if history_item.model_output else ''
			logger.info(f'Replaying step {i + 1}/{len(history.history)}: goal: {goal}')
			
			if not history_item.model_output or not history_item.model_output.action or history_item.model_output.action == [None]:
				logger.warning(f'Step {i + 1}: No action to replay, skipping')
				results.append(ActionResult(error='No action to replay'))
				continue
				
			retry_count = 0
			while retry_count < max_retries:
				try:
					result = await self._execute_history_step(history_item, delay_between_actions)
					results.extend(result)
					break
				except Exception as e:
					retry_count += 1
					if retry_count == max_retries:
						error_msg = f'Step {i + 1} failed after {max_retries} attempts: {str(e)}'
						logger.error(error_msg)
						if not skip_failures:
							results.append(ActionResult(error=error_msg))
							raise RuntimeError(error_msg)
					else:
						logger.warning(f'Step {i + 1} failed (attempt {retry_count}/{max_retries}), retrying...')
						await asyncio.sleep(delay_between_actions)
												
		return results
								
	async def _execute_history_step(self, history_item: AgentHistory, delay: float) -> list[ActionResult]:
		"""Execute a single step from history with element validation"""
		# Get current browser state
		state = await self.browser_context.get_state()
		
		if not state or not history_item.model_output:
			raise ValueError('Invalid state or model output')
			
		updated_actions = []
		
		# Update action indices based on current DOM structure
		for i, action in enumerate(history_item.model_output.action):
			updated_action = await self._update_action_indices(
				history_item.state.interacted_element[i],
				action,
				state
			)
			updated_actions.append(updated_action)
			
			if updated_action is None:
				raise ValueError(f'Could not find matching element {i} in current page')
				
		# Execute updated actions
		result = await self.controller.multi_act(
			updated_actions,
			self.browser_context,
			self.page_extraction_llm,
			check_break_if_paused=lambda: self._check_if_stopped_or_paused()
		)
		
		await asyncio.sleep(delay)
		return result
		
	async def _update_action_indices(
		self,
		historical_element: Optional[DOMHistoryElement],
		action: ActionModel, 
		current_state: BrowserState
	) -> Optional[ActionModel]:
		"""
		Update action indices based on current page state.
		Returns updated action or None if element cannot be found.
		"""
		if not historical_element or not current_state.element_tree:
			return action
			
		current_element = HistoryTreeProcessor.find_history_element_in_tree(
			historical_element, 
			current_state.element_tree
		)
		
		if not current_element or current_element.highlight_index is None:
			return None
			
		old_index = action.get_index()
		if old_index != current_element.highlight_index:
			action.set_index(current_element.highlight_index)
			logger.info(f'Element moved in DOM, updated index from {old_index} to {current_element.highlight_index}')
			
		return action
		
	async def load_and_rerun(self, history_file: Optional[str | Path] = None, **kwargs) -> list[ActionResult]:
		"""
		Load history from file and rerun it.

		Args:
			history_file: Path to the history file
			**kwargs: Additional arguments passed to rerun_history
		"""
		if not history_file:
			history_file = 'AgentHistory.json'
			
		history = AgentHistoryList.load_from_file(history_file, self.AgentOutput)
		return await self.rerun_history(history, **kwargs)
		
	def save_history(self, file_path: Optional[str | Path] = None) -> None:
		"""Save the history to a file"""
		if not file_path:
			file_path = 'AgentHistory.json'
			
		self.history.save_to_file(file_path)
		
	def create_history_gif(
		self,
		output_path: str = 'agent_history.gif',
		duration: int = 3000,
		show_goals: bool = True,
		show_task: bool = True,
		show_logo: bool = False,
		font_size: int = 40,
		title_font_size: int = 56,
		goal_font_size: int = 44,
		margin: int = 40,
		line_spacing: float = 1.5
	) -> None:
		"""Create a GIF from the agent's history with overlaid task and goal text."""
		if not self.history.history:
			logger.warning('No history to create GIF from')
			return None
			
		frames = []
		
		if not self.history.history or not self.history.history[0].state.screenshot:
			logger.warning('No history or first screenshot to create GIF from')
			return None
			
		# Try to find available fonts
		font_names = ['Helvetica', 'Arial', 'DejaVuSans', 'Verdana']
		font_found = False
		
		for font_name in font_names:
			try:
				if platform.system() == 'Windows':
					font_name = os.path.join(os.getenv('WIN_FONT_DIR', 'C:\\Windows\\Fonts'), font_name + '.ttf')
					
				regular_font = ImageFont.truetype(font_name, font_size)
				title_font = ImageFont.truetype(font_name, title_font_size)
				goal_font = ImageFont.truetype(font_name, goal_font_size)
				font_found = True
				break
			except OSError:
				continue
				
		if not font_found:
			try:
				regular_font = ImageFont.load_default()
				title_font = ImageFont.load_default()
				goal_font = ImageFont.load_default()
			except OSError:
				raise OSError('No preferred fonts found')
				
		# Try to load logo if enabled
		logo_image = None
		if show_logo:
			try:
				logo_path = os.path.join(os.path.dirname(__file__), 'assets/browser-use-logo.png')
				if os.path.exists(logo_path):
					logo_image = Image.open(logo_path)
					logo_image.thumbnail((150, 150))
			except Exception as e:
				logger.warning(f'Could not load logo: {e}')
				
		# Create task frame if enabled
		if show_task and self.task:
			task_frame = self._create_task_frame(
				self.task,
				self.history.history[0].state.screenshot,
				title_font,
				regular_font,
				logo_image,
				line_spacing
			)
			frames.append(task_frame)
			
		# Create frames for each step
		for i, history_item in enumerate(self.history.history[1:], 1):
			if not history_item.state.screenshot:
				continue
				
			# Decode screenshot
			image_data = base64.b64decode(history_item.state.screenshot)
			
			# Open image
			img = Image.open(io.BytesIO(image_data))
			
			# Add text overlay if goals are enabled
			if show_goals and history_item.model_output:
				img = self._add_overlay_to_image(
					image=img,
					step_number=i,
					goal_text=history_item.model_output.current_state.next_goal,
					regular_font=regular_font,
					title_font=title_font,
					margin=margin,
					logo=logo_image
				)
				
			frames.append(img)
			
		# Save as GIF
		if frames:
			frames[0].save(
				output_path,
				save_all=True,
				append_images=frames[1:],
				duration=duration,
				loop=0,
				optimize=False
			)
			logger.info(f'Created GIF at {output_path}')
		else:
			logger.warning('No images found in history to create GIF')
			
	def _create_task_frame(
		self,
		task: str,
		first_screenshot: str,
		title_font,
		regular_font,
		logo=None,
		line_spacing=1.5
	) -> Image.Image:
		"""Create initial frame showing the task."""
		# Decode screenshot
		screenshot_data = base64.b64decode(first_screenshot)
		screenshot_image = Image.open(io.BytesIO(screenshot_data))
		
		# Create black background
		background_image = Image.new('RGB', screenshot_image.size, (0, 0, 0))
		draw = ImageDraw.Draw(background_image)
		
		# Position variables
		center_y = background_image.height // 2
		margin_x = 140
		text_width = background_image.width - 2 * margin_x
		
		# Create larger font for task title
		task_title_font = ImageFont.truetype(regular_font.path, regular_font.size + 16)
		
		# Wrap text to fit width
		wrapped_text = self._wrap_text(task, task_title_font, text_width)
		line_height = task_title_font.size * line_spacing
		lines = wrapped_text.split('\n')
		total_text_height = line_height * len(lines)
		text_start_y = (center_y - total_text_height / 2) + 50
		
		# Draw each line of the task
		for line in lines:
			# Calculate text width for centering
			text_bbox = draw.textbbox((0, 0), line, font=task_title_font)
			text_start_x = (background_image.width - (text_bbox[2] - text_bbox[0])) // 2
			
			# Draw text
			draw.text(
				(text_start_x, text_start_y),
				line,
				font=task_title_font,
				fill=(255, 255, 255)
			)
			
			text_start_y += line_height
			
		# Add logo if available
		if logo:
			logo_margin = 20
			logo_x = background_image.width - logo.width - logo_margin
			background_image.paste(
				logo,
				(logo_x, logo_margin),
				logo if 'A' in logo.getbands() else None
			)
			
		return background_image
		
	def _add_overlay_to_image(
		self,
		image,
		step_number,
		goal_text,
		regular_font,
		title_font,
		margin,
		logo=None,
		display_step=True,
		text_color=(255, 255, 255, 255),
		text_box_color=(0, 0, 0, 255)
	) -> Image.Image:
		"""Add step number and goal overlay to an image."""
		# Convert to RGBA for transparency
		image = image.convert('RGBA')
		overlay = Image.new('RGBA', image.size, (0, 0, 0, 0))
		draw = ImageDraw.Draw(overlay)
		
		# Add step number if enabled
		if display_step:
			step_text = str(step_number)
			step_bbox = draw.textbbox((0, 0), step_text, font=title_font)
			step_width = step_bbox[2] - step_bbox[0]
			step_height = step_bbox[3] - step_bbox[1]
			
			# Position for step number
			step_x = margin + 10
			step_y = image.height - margin - step_height - 10
			padding = 20
			
			# Background for step number
			step_background_bbox = (
				step_x - padding,
				step_y - padding,
				step_x + step_width + padding,
				step_y + step_height + padding
			)
			
			draw.rounded_rectangle(step_background_bbox, radius=15, fill=text_box_color)
			draw.text((step_x, step_y), step_text, font=title_font, fill=text_color)
			
		# Add goal text
		text_width = image.width - 4 * margin
		wrapped_text = self._wrap_text(goal_text, title_font, text_width)
		
		# Get text dimensions
		text_bbox = draw.multiline_textbbox((0, 0), wrapped_text, font=title_font)
		text_w = text_bbox[2] - text_bbox[0]
		text_h = text_bbox[3] - text_bbox[1]
		
		# Position for goal text
		text_x = (image.width - text_w) // 2
		text_y = step_y - text_h - padding * 4  # Use variables from step number section
		text_padding = 25
		
		# Background for goal text
		text_background_bbox = (
			text_x - text_padding,
			text_y - text_padding,
			text_x + text_w + text_padding,
			text_y + text_h + text_padding
		)
		
		draw.rounded_rectangle(text_background_bbox, radius=15, fill=text_box_color)
		draw.multiline_text(
			(text_x, text_y),
			wrapped_text,
			font=title_font,
			fill=text_color,
			align='center'
		)
		
		# Add logo if available
		if logo:
			logo_overlay = Image.new('RGBA', image.size, (0, 0, 0, 0))
			logo_margin = 20
			logo_x = image.width - logo.width - logo_margin
			logo_overlay.paste(
				logo,
				(logo_x, logo_margin),
				logo if logo.mode == 'RGBA' else None
			)
			overlay = Image.alpha_composite(logo_overlay, overlay)
			
		# Composite the overlay with the original image
		composite_image = Image.alpha_composite(image, overlay)
		return composite_image.convert('RGB')
		
	def _wrap_text(self, text: str, font, max_width: int) -> str:
		"""
		Wrap text to fit within a given width.

		Args:
			text: Text to wrap
			font: Font to use for text
			max_width: Maximum width in pixels

		Returns:
			Wrapped text with newlines
		"""
		words = text.split()
		lines = []
		current_line = []
		
		for word in words:
			current_line.append(word)
			line_text = ' '.join(current_line)
			line_bbox = font.getbbox(line_text)
			
			if line_bbox[2] > max_width:
				if len(current_line) == 1:
					# Single word is too long, just add it anyway
					lines.append(current_line.pop())
				else:
					# Remove last word and add the rest to wrapped lines
					current_line.pop()
					lines.append(' '.join(current_line))
					current_line = [word]
					
		if current_line:
			lines.append(' '.join(current_line))
			
		return '\n'.join(lines)
		
	def _create_frame(
		self,
		screenshot: str,
		text: str,
		step_number: int,
		width: int = 1200,
		height: int = 800
	) -> Image.Image:
		"""Create a frame for the GIF with improved styling"""
		# Create base image
		base_image = Image.new('RGB', (width, height), 'white')
		
		# Load screenshot
		screenshot_image = Image.open(BytesIO(base64.b64decode(screenshot)))
		screenshot_image.thumbnail((width - 40, height - 160))
		
		# Center screenshot
		screenshot_x = (width - screenshot_image.width) // 2
		screenshot_y = 120
		base_image.paste(screenshot_image, (screenshot_x, screenshot_y))
		
		# Try to load logo
		logo_size = 100
		logo_path = os.path.join(os.path.dirname(__file__), 'assets/browser-use-logo.png')
		
		if os.path.exists(logo_path):
			logo_image = Image.open(logo_path)
			logo_image.thumbnail((logo_size, logo_size))
			base_image.paste(
				logo_image,
				(width - logo_size - 20, 20),
				logo_image if 'A' in logo_image.getbands() else None
			)
			
		# Setup for drawing text
		draw = ImageDraw.Draw(base_image)
		
		try:
			title_font = ImageFont.truetype('Arial.ttf', 36)
			regular_font = ImageFont.truetype('Arial.ttf', 24)
			step_number_font = ImageFont.truetype('Arial.ttf', 48)
		except:
			# Fallback to default fonts
			title_font = ImageFont.load_default()
			regular_font = ImageFont.load_default()
			step_number_font = ImageFont.load_default()
			
		# Text positioning
		margin_x = 80
		text_width = width - 2 * margin_x
		padding = 20
		
		# Wrap text
		wrapped_text = textwrap.wrap(text, width=60)
		
		# Calculate text height
		total_text_height = sum(
			draw.textsize(line, font=regular_font)[1]
			for line in wrapped_text
		)
		
		text_area_height = total_text_height + 2 * padding
		
		# Create text area background
		text_area_bbox = [
			margin_x - padding,
			40,
			(width - margin_x) + padding,
			40 + text_area_height
		]
		
		draw.rounded_rectangle(text_area_bbox, radius=15, fill='#f0f0f0')
		
		# Add small logo to text area
		small_logo_size = 30
		if os.path.exists(logo_path):
			small_logo_image = Image.open(logo_path)
			small_logo_image.thumbnail((small_logo_size, small_logo_size))
			base_image.paste(
				small_logo_image,
				((margin_x - padding) + 10, 45),
				small_logo_image if 'A' in small_logo_image.getbands() else None
			)
			
		# Draw text
		text_y = 50
		for line in wrapped_text:
			draw.text(
				(margin_x + small_logo_size + 20, text_y),
				line,
				font=regular_font,
				fill='black'
			)
			text_y += draw.textsize(line, font=regular_font)[1] + 5
			
		# Add step number
		step_number_text = str(step_number)
		step_number_bbox = draw.textsize(step_number_text, font=step_number_font)
		step_number_padding = 20
		step_number_width = step_number_bbox[0] + 2 * step_number_padding
		step_number_height = step_number_bbox[1] + 2 * step_number_padding
		
		# Step number background
		step_number_background_bbox = [
			20,
			height - step_number_height - 20,
			20 + step_number_width,
			height - 20
		]
		
		draw.rounded_rectangle(step_number_background_bbox, radius=15, fill='#007AFF')
		
		# Calculate center position for step number
		step_number_x = step_number_background_bbox[0] + (step_number_width - step_number_bbox[0]) // 2
		step_number_y = step_number_background_bbox[1] + (step_number_height - step_number_bbox[1]) // 2
		
		# Draw step number
		draw.text(
			(step_number_x, step_number_y),
			step_number_text,
			font=step_number_font,
			fill='white'
		)
		
		return base_image
		
	def pause(self):
		"""Pause the agent before the next step"""
		logger.info('🔄 pausing Agent ')
		self._paused = True
		
	def resume(self):
		"""Resume the agent"""
		logger.info('▶️ Agent resuming')
		self._paused = False
		
	def stop(self):
		"""Stop the agent"""
		logger.info('⏹️ Agent stopping')
		self._stopped = True
		
	def _convert_initial_actions(self, actions: List[Dict[str, Dict[str, Any]]]) -> List[ActionModel]:
		"""Convert dictionary-based actions to ActionModel instances"""
		converted_actions = []
		action_model_instance = self.ActionModel  # Renamed for clarity
		for action_dict in actions:
			# Each action_dict should have a single key-value pair
			action_name = next(iter(action_dict))
			params = action_dict[action_name]

			# Get the parameter model for this action from registry
			action_info = self.controller.registry.registry.actions[action_name]
			param_model = action_info.param_model

			# Create validated parameters using the appropriate param model
			validated_params = param_model(**params)

			# Create ActionModel instance with the validated parameters
			action_model_instance = self.ActionModel(**{action_name: validated_params}) # Use the renamed variable
			converted_actions.append(action_model_instance)

		return converted_actions

	async def _run_planner(self) -> Optional[str]:
		"""Run the planner to analyze state and suggest next steps"""
		# Skip planning if no planner_llm is set
		if not self.planner_llm:
			return None

		# Create planner message history using full message history
		planner_messages = [
			PlannerPrompt(self.controller.registry.get_prompt_description()).get_system_message(),
			*self.message_manager.get_messages()[1:],  # Use full message history except the first
		]

		if not self.use_vision_for_planner and self.use_vision:
			last_state_message: HumanMessage = planner_messages[-1]
			# remove image from last state message
			new_message_content = ''
			if isinstance(last_state_message.content, list):
				for message_part in last_state_message.content:
					if message_part['type'] == 'text':  # type: ignore
						new_message_content += message_part['text']  # type: ignore
					elif message_part['type'] == 'image_url':  # type: ignore
						continue  # type: ignore
			else:
				new_message_content = last_state_message.content

			planner_messages[-1] = HumanMessage(content=new_message_content)

		planner_messages = self._convert_input_messages(planner_messages, self.planner_model_name)

		# Get planner output
		response = await self.planner_llm.ainvoke(planner_messages)
		plan = str(response.content)
		# if deepseek-reasoner, remove think tags
		if self.planner_model_name and ('deepseek-r1' in self.planner_model_name or 'deepseek-reasoner' in self.planner_model_name):
			plan = self._remove_think_tags(plan)
		try:
			plan_json = json.loads(plan)
			logger.info(f'Planning Analysis:\n{json.dumps(plan_json, indent=4)}')
		except json.JSONDecodeError:
			logger.info(f'Planning Analysis:\n{plan}')
		except Exception as error:
			logger.debug(f'Error parsing planning analysis: {error}')
			logger.info(f'Plan: {plan}')

		return plan

	@property
	def message_manager(self) -> MessageManager:
		return self._message_manager


"""
Orion Browser: 浏览器代理服务器包

该包提供了浏览器代理功能，可以用作独立服务或导入为Python包使用。
"""

# 直接从app模块导入BrowserManager
from app.tools.browser.browser_manager import BrowserManager
from app.types.messages import BrowserActionRequest
from app.types.browser_types import (
    BrowserAction, GoToUrlAction, NoParamAction, ScrollToTextAction,
    GetDropdownOptionsAction, SelectDropdownOptionAction, ViewAction,
    DoneAction, SaveImageAction, SaveScreenshotAction, ExtractPageContentAction,
    BrowserNavigateAction, BrowserViewAction, BrowserScreenshotAction,
    BrowserRestartAction, BrowserClickAction, BrowserMoveMouseAction,
    BrowserInputAction, BrowserPressKeyAction, BrowserScrollUpAction,
    BrowserScrollDownAction, BrowserSelectOptionAction, BrowserConsoleExecAction,
    BrowserConsoleViewAction, SearchBaiduAction, ClickElementAction, InputTextAction,
    ClickByPositionAction,OpenTabAction, ScrollAction, SwitchTabAction, SendKeysAction, BrowserActionResult,
    GetAllTabsAction
)

# 导出核心组件和所有 action 类型
__all__ = [
    "BrowserManager", 
    "BrowserActionRequest",
    "BrowserAction", 
    "GoToUrlAction",
    "NoParamAction",
    "ScrollToTextAction",
    "GetDropdownOptionsAction", 
    "SelectDropdownOptionAction", 
    "ViewAction",
    "DoneAction", 
    "SaveImageAction", 
    "SaveScreenshotAction", 
    "ExtractPageContentAction",
    "BrowserNavigateAction", 
    "BrowserViewAction", 
    "BrowserScreenshotAction",
    "BrowserRestartAction", 
    "BrowserClickAction", 
    "BrowserMoveMouseAction",
    "BrowserInputAction", 
    "BrowserPressKeyAction", 
    "BrowserScrollUpAction",
    "BrowserScrollDownAction", 
    "BrowserSelectOptionAction", 
    "BrowserConsoleExecAction",
    "BrowserConsoleViewAction", 
    "SearchBaiduAction", 
    "ClickElementAction", 
    "ClickByPositionAction",
    "InputTextAction",
    "OpenTabAction", 
    "GetAllTabsAction",
    "ScrollAction", 
    "SwitchTabAction", 
    "SendKeysAction",
    "BrowserActionResult"
] 


import asyncio
import json
from fastapi import WebSocket, WebSocketDisconnect
from pydantic import ValidationError
from app.logger import logger
from app.tools.terminal.terminal_manager import terminal_manager
from app.types.messages import TerminalInputMessage, TerminalOutputMessage, TerminalStatus

class TerminalSocket:
    """
    WebSocket server for handling terminal connections.
    This class manages bidirectional communication with terminals through WebSockets.
    """
    
    async def handle_connection(self, ws: WebSocket):
        """
        Handle a new WebSocket connection for terminal interaction.
        
        Args:
            ws: The WebSocket connection
        """
        await ws.accept()
        logger.info("New terminal WebSocket connection established")
        
        tasks = {}
        def stop_all_tasks():
            for task in tasks.values():
                task.cancel()
        
        async def get_socket_message():
            try:
                msg_data = await ws.receive_json()
                msg = TerminalInputMessage.model_validate(msg_data)
                task = asyncio.create_task(self.handle_msg(ws, msg))
                tasks[msg.action_id] = task
                task.add_done_callback(lambda _: tasks.pop(msg.action_id))
            except ValidationError as e:
                logger.error(f"Invalid message: {msg_data}, {e}")
                await self.send_resp(ws, TerminalOutputMessage(
                    action_id="",
                    type="error",
                    result=f"Invalid message: {e}",
                    output=[],
                    terminal_status="unknown",
                    terminal=""
                ))
            except json.JSONDecodeError as e:
                logger.error(f"JSONDecodeError: {e}")
                await self.send_resp(ws, TerminalOutputMessage(
                    action_id="",
                    type="error",
                    result=f"Invalid message: {e}",
                    output=[],
                    terminal_status="unknown",
                    terminal=""
                ))
            except Exception as e:
                logger.error(f"Error handling message: {e}")
                raise
        
        try:
            while True:
                await get_socket_message()
        except WebSocketDisconnect:
            logger.info("Websocket disconnected")
        except Exception as e:
            logger.error(e)
            logger.error(f"Error: {e}")
            logger.info("Closing websocket")
            await ws.close()
        
        stop_all_tasks()
    
    async def send_resp(self, ws: WebSocket, resp: TerminalOutputMessage):
        """
        Send a response to the client over the WebSocket connection.
        
        Args:
            ws: The WebSocket connection
            resp: The response data to send
        """
        logger.info(f"Sending resp {resp}")
        
        try:
            await ws.send_json(resp.model_dump())
        except RuntimeError as e:
            logger.error(f"Error sending resp: {e}")
    
    async def handle_msg(self, ws: WebSocket, msg: TerminalInputMessage):
        """
        Process a validated terminal input message.
        
        Args:
            msg: The validated terminal input message
            ws: The WebSocket connection
        """
        logger.info(f"Handle terminal socket msg#{msg.action_id} {msg}")
        
        terminal_id = msg.terminal
        terminal = await terminal_manager.get_terminal(terminal_id)
        
        if msg.type == "reset":
            await terminal.reset()
            response = msg.create_response(
                type="action_finish",
                result="terminal reset success",
                output=[],
                terminal_status="idle"
            )
            await self.send_resp(ws, response)
        
        elif msg.type == "reset_all":
            for term in terminal_manager.terminals.values():
                await term.reset()
            response = msg.create_response(
                type="action_finish",
                result="all terminals reset success",
                output=[],
                terminal_status="idle"
            )
            await self.send_resp(ws, response)
        
        elif msg.type == "view":
            terminal_status: TerminalStatus = "running" if terminal.is_running else "idle"
            response = msg.create_response(
                type="history",
                result=None,
                output=terminal.get_history(True, True),
                terminal_status=terminal_status
            )
            await self.send_resp(ws, response)
        
        elif msg.type == "view_last":
            terminal_status: TerminalStatus = "running" if terminal.is_running else "idle"
            response = msg.create_response(
                type="history",
                result=None,
                output=terminal.get_history(True, False),
                terminal_status=terminal_status
            )
            await self.send_resp(ws, response)
        
        elif msg.type == "kill":
            await terminal.kill_process()
            response = msg.create_response(
                type="action_finish",
                result="process killed",
                output=terminal.get_history(True, False),
                terminal_status="idle"
            )
            await self.send_resp(ws, response)
        
        elif msg.type == "command":
            if not msg.command:
                response = msg.create_response(
                    type="error",
                    result="must provide command",
                    output=[],
                    terminal_status="idle"
                )
                await self.send_resp(ws, response)
                return
            
            if msg.exec_dir:
                if not await terminal.set_working_directory(msg.exec_dir):
                    response = msg.create_response(
                        type="error",
                        result=f"Failed to change directory to {msg.exec_dir}",
                        output=[],
                        terminal_status="idle"
                    )
                    await self.send_resp(ws, response)
                    return
            
            if not msg.mode:
                msg.mode = "run"
            
            if msg.mode == "send_key":
                await terminal.send_key(msg)
                terminal_status: TerminalStatus = "idle" if not terminal.is_running else "running"
                response = msg.create_response(
                    type="action_finish",
                    result=f"Key sent: {msg.command}",
                    output=terminal.get_history(True, False),
                    terminal_status=terminal_status
                )
                await self.send_resp(ws, response)
            
            elif msg.mode == "send_line":
                await terminal.send_line(msg)
                terminal_status: TerminalStatus = "idle" if not terminal.is_running else "running"
                response = msg.create_response(
                    type="action_finish",
                    result=f"Line sent: {msg.command}",
                    output=terminal.get_history(True, False),
                    terminal_status=terminal_status
                )
                await self.send_resp(ws, response)
            
            elif msg.mode == "send_control":
                await terminal.send_control(msg)
                terminal_status: TerminalStatus = "idle" if not terminal.is_running else "running"
                response = msg.create_response(
                    type="action_finish",
                    result=f"Control character sent: {msg.command}",
                    output=terminal.get_history(True, False),
                    terminal_status=terminal_status
                )
                await self.send_resp(ws, response)
            
            elif msg.mode == "run":
                async for result in terminal.execute_command(msg):
                    await self.send_resp(ws, result)
            
            else:
                response = msg.create_response(
                    type="error",
                    result=f"Invalid mode: {msg.mode}",
                    output=[],
                    terminal_status="idle"
                )
                await self.send_resp(ws, response)
        
        else:
            logger.error(f"Invalid message type: {msg.type}")
            response = msg.create_response(
                type="error",
                result=f"Invalid message type: {msg.type}",
                output=[],
                terminal_status="idle"
            )
            await self.send_resp(ws, response)
        
        logger.info(f"Finished handling msg#{msg.action_id}")

terminal_socket = TerminalSocket()


from __future__ import annotations
from typing import TYPE_CHECKING, Any, Optional
from langchain_core.load import dumpd, load
from langchain_core.messages import AIMessage, BaseMessage, HumanMessage, SystemMessage, ToolMessage
from pydantic import BaseModel, ConfigDict, Field, model_serializer, model_validator

if TYPE_CHECKING:
    from browser_use.agent.views import AgentOutput


class MessageMetadata(BaseModel):
    """Metadata for a message including token counts"""
    tokens: int = 0
    input_tokens: int = 0  # For backward compatibility

    model_config = ConfigDict(arbitrary_types_allowed=True)


class ManagedMessage(BaseModel):
    """A message with its metadata"""
    message: BaseMessage
    metadata: MessageMetadata = Field(default_factory=MessageMetadata)
    
    model_config = ConfigDict(arbitrary_types_allowed=True)
    
    @model_serializer(mode='wrap')
    def to_json(self, original_dump):
        """
        Return the JSON representation of the model.
        Uses langchain's `dumpd` to serialize the message field.
        """
        data = original_dump(self)
        data['message'] = dumpd(self.message)
        return data
    
    @model_validator(mode='before')
    @classmethod
    def validate(
        cls,
        value: Any,
        *,
        strict: bool | None = None,
        from_attributes: bool | None = None,
        context: Any | None = None,
    ) -> Any:
        """
        Custom validator that uses langchain's `load` function
        to parse the message if it is provided as a JSON string.
        """
        if isinstance(value, dict) and 'message' in value:
            value['message'] = load(value['message'])
        return value


class MessageHistory(BaseModel):
    """Container for message history with metadata"""
    messages: list[ManagedMessage] = Field(default_factory=list)
    current_tokens: int = 0
    
    model_config = ConfigDict(arbitrary_types_allowed=True)
    
    def add_message(self, message: BaseMessage, metadata: MessageMetadata, position: Optional[int] = None) -> None:
        """
        Add a message with its metadata to history.

        Args:
            message: The message to add.
            metadata: Metadata including token counts.
            position: Optional index at which to insert the message (None to append).
        """
        if position is None:
            self.messages.append(ManagedMessage(message=message, metadata=metadata))
        else:
            self.messages.insert(position, ManagedMessage(message=message, metadata=metadata))
        
        token_count = metadata.tokens or metadata.input_tokens
        self.current_tokens += token_count
    
    def add_model_output(self, output: 'AgentOutput') -> None:
        """
        Add model output as an AI message with an associated tool call.
        """
        tool_calls = [
            {
                'name': 'AgentOutput',
                'args': output.model_dump(mode='json', exclude_unset=True),
                'id': '1',
                'type': 'tool_call',
            }
        ]
        
        msg = AIMessage(
            content='',
            tool_calls=tool_calls,
        )
        
        # Estimate tokens for tool call message
        self.add_message(msg, MessageMetadata(tokens=100))
        
        # Add an empty tool response (estimated token count)
        tool_message = ToolMessage(content='', tool_call_id='1')
        self.add_message(tool_message, MessageMetadata(tokens=10))
    
    def get_messages(self) -> list[BaseMessage]:
        """Return a list of all messages (without metadata)."""
        return [m.message for m in self.messages]
    
    def get_total_tokens(self) -> int:
        """Return the total token count for the message history."""
        return self.current_tokens
    
    def remove_message(self, index: int = -1) -> None:
        """
        Remove a message at the specified index (default: last message)
        and update the total token count.
        """
        if 0 <= abs(index) < len(self.messages):
            msg = self.messages.pop(index)
            token_count = msg.metadata.tokens or msg.metadata.input_tokens
            self.current_tokens -= token_count
    
    def remove_oldest_message(self) -> None:
        """
        Remove the oldest non-system message from history.
        """
        for i, msg in enumerate(self.messages):
            if not isinstance(msg.message, SystemMessage):
                token_count = msg.metadata.tokens or msg.metadata.input_tokens
                self.current_tokens -= token_count
                self.messages.pop(i)
                break
    
    def remove_last_state_message(self) -> None:
        """
        Remove the last state (Human) message from history if there are more than 2 messages.
        """
        if len(self.messages) > 2 and isinstance(self.messages[-1].message, HumanMessage):
            token_count = self.messages[-1].metadata.tokens or self.messages[-1].metadata.input_tokens
            self.current_tokens -= token_count
            self.messages.pop()
    
    @property
    def total_tokens(self) -> int:
        """Backward compatibility: total tokens in history."""
        return self.current_tokens


class MessageManagerState(BaseModel):
    """Holds the state for MessageManager"""
    history: MessageHistory = Field(default_factory=MessageHistory)
    tool_id: int = 1
    
    model_config = ConfigDict(arbitrary_types_allowed=True)



import os
import pathlib
import base64
from dataclasses import dataclass

from app.logger import logger
from browser_use import ActionResult

__all__ = [
    'HelperJs',
    'screenshot_to_data_url',
    'check_file_path'
]

# Directory where JavaScript helper files are stored
JS_DIR = os.path.normpath(os.path.join(os.path.dirname(__file__), '../../../app_data/js'))

@dataclass
class HelperJs:
    """Container for JavaScript helper code snippets used for browser automation."""
    
    # Extract page content helper
    EXTRACT_CONTENT = """
    function extractPageContent() {
        // Get title and metadata
        const title = document.title;
        const metaDescription = document.querySelector('meta[name="description"]')?.content || '';
        
        // Get main content
        let mainContent = '';
        const mainElement = document.querySelector('main') || document.body;
        mainContent = mainElement.innerText;
        
        // Get all links
        const links = [];
        document.querySelectorAll('a[href]').forEach(link => {
            links.push({
                text: link.innerText.trim(),
                href: link.href
            });
        });
        
        return {
            title,
            metaDescription,
            mainContent,
            links,
            url: window.location.href
        };
    }
    
    return JSON.stringify(extractPageContent());
    """
    
    # Find clickable elements helper
    FIND_CLICKABLE = """
    function findClickableElements() {
        const clickableElements = [];
        
        // Common clickable elements
        const selectors = [
            'a', 'button', 'input[type="button"]', 'input[type="submit"]',
            '.btn', '[role="button"]', '[onclick]', 'select', 'summary',
            'details', '[tabindex]:not([tabindex="-1"])'
        ];
        
        const elements = document.querySelectorAll(selectors.join(', '));
        
        elements.forEach((el, index) => {
            // Check if element is visible
            const rect = el.getBoundingClientRect();
            const isVisible = !!(rect.width && rect.height &&
                window.getComputedStyle(el).getPropertyValue('display') !== 'none' &&
                window.getComputedStyle(el).getPropertyValue('visibility') !== 'hidden');
                
            if (isVisible) {
                clickableElements.push({
                    index,
                    tagName: el.tagName.toLowerCase(),
                    id: el.id || '',
                    className: el.className || '',
                    text: el.innerText?.trim() || el.textContent?.trim() || '',
                    name: el.name || '',
                    href: el.href || '',
                    type: el.type || '',
                    value: el.value || '',
                    ariaLabel: el.getAttribute('aria-label') || '',
                    rect: {
                        x: rect.x,
                        y: rect.y,
                        width: rect.width,
                        height: rect.height
                    }
                });
            }
        });
        
        return clickableElements;
    }
    
    return JSON.stringify(findClickableElements());
    """
    
    # Find input elements helper
    FIND_INPUTS = """
    function findInputElements() {
        const inputElements = [];
        
        // Input selectors
        const selectors = [
            'input:not([type="button"]):not([type="submit"]):not([type="reset"]):not([type="hidden"])',
            'textarea',
            '[contenteditable="true"]',
            '[role="textbox"]'
        ];
        
        const elements = document.querySelectorAll(selectors.join(', '));
        
        elements.forEach((el, index) => {
            // Check if element is visible
            const rect = el.getBoundingClientRect();
            const isVisible = !!(rect.width && rect.height &&
                window.getComputedStyle(el).getPropertyValue('display') !== 'none' &&
                window.getComputedStyle(el).getPropertyValue('visibility') !== 'hidden');
                
            if (isVisible) {
                inputElements.push({
                    index,
                    tagName: el.tagName.toLowerCase(),
                    id: el.id || '',
                    className: el.className || '',
                    name: el.name || '',
                    type: el.type || '',
                    placeholder: el.placeholder || '',
                    value: el.value || '',
                    ariaLabel: el.getAttribute('aria-label') || '',
                    rect: {
                        x: rect.x,
                        y: rect.y,
                        width: rect.width,
                        height: rect.height
                    }
                });
            }
        });
        
        return inputElements;
    }
    
    return JSON.stringify(findInputElements());
    """
    
    # Console log helper
    CONSOLE_LOGS = """
    function getConsoleLogs(maxLines) {
        if (!window.__consoleLogs) {
            return "Console logging not initialized. Refresh the page to enable logging.";
        }
        
        const logs = window.__consoleLogs;
        if (maxLines && logs.length > maxLines) {
            return logs.slice(logs.length - maxLines).join("\\n");
        }
        
        return logs.join("\\n");
    }
    
    return getConsoleLogs(%d);
    """
    
    # Initialize console logging
    INIT_CONSOLE_LOGGING = """
    (function initializeConsoleLogging() {
        if (window.__consoleLogs) {
            return "Console logging already initialized.";
        }
        
        window.__consoleLogs = [];
        
        // Store original console methods
        const originalConsole = {
            log: console.log,
            info: console.info,
            warn: console.warn,
            error: console.error,
            debug: console.debug
        };
        
        // Override console methods to capture logs
        console.log = function() {
            window.__consoleLogs.push(["LOG", ...arguments].map(arg =>
                typeof arg === 'object' ? JSON.stringify(arg) : String(arg)).join(" "));
            return originalConsole.log.apply(console, arguments);
        };
        
        console.info = function() {
            window.__consoleLogs.push(["INFO", ...arguments].map(arg =>
                typeof arg === 'object' ? JSON.stringify(arg) : String(arg)).join(" "));
            return originalConsole.info.apply(console, arguments);
        };
        
        console.warn = function() {
            window.__consoleLogs.push(["WARN", ...arguments].map(arg =>
                typeof arg === 'object' ? JSON.stringify(arg) : String(arg)).join(" "));
            return originalConsole.warn.apply(console, arguments);
        };
        
        console.error = function() {
            window.__consoleLogs.push(["ERROR", ...arguments].map(arg =>
                typeof arg === 'object' ? JSON.stringify(arg) : String(arg)).join(" "));
            return originalConsole.error.apply(console, arguments);
        };
        
        console.debug = function() {
            window.__consoleLogs.push(["DEBUG", ...arguments].map(arg =>
                typeof arg === 'object' ? JSON.stringify(arg) : String(arg)).join(" "));
            return originalConsole.debug.apply(console, arguments);
        };
        
        return "Console logging initialized.";
    })();
    """

# Read additional JavaScript helpers from files in JS_DIR
try:
    select_option = pathlib.Path(JS_DIR, "selectOption.js").read_text()
except Exception as e:
    logger.error(f"Error reading selectOption.js: {e}")
    select_option = ""

try:
    get_viewport = pathlib.Path(JS_DIR, "getViewport.js").read_text()
except Exception as e:
    logger.error(f"Error reading getViewport.js: {e}")
    get_viewport = ""

def screenshot_to_data_url(screenshot):
    '''将 base64 编码的截图转换为 data URL 格式 (Convert base64 encoded screenshot to data URL format)'''
    if isinstance(screenshot, bytes):
        screenshot = base64.b64encode(screenshot).decode('utf-8')
    return f'data:image/png;base64,{screenshot}'

def check_file_path(file_path):
    '''
    检查文件路径并创建必要的目录 (Check file path and create necessary directories)
    
    Args:
        file_path: 文件路径，必须是绝对路径 (file path, must be an absolute path)
        
    Returns:
        如果有错误返回 ActionResult，否则返回 None (Return ActionResult if there is an error, otherwise return None)
        
    错误情况: (Error cases:)
        - file_path 为空 (file_path is empty)
        - file_path 不是绝对路径 (file_path is not an absolute path)
        - file_path 指向的文件已存在 (file pointed to by file_path already exists)
        - 无法创建目录 (unable to create directory)
    '''
    if not file_path:
        return ActionResult(error='file_path is required', include_in_memory=True)
        
    if not file_path.startswith('/'):
        return ActionResult(error=f'File path must be absolute: {file_path}', include_in_memory=True)
        
    if pathlib.Path(file_path).exists():
        return ActionResult(error=f'File already exists: {file_path}', include_in_memory=True)
        
    try:
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        return None
    except Exception as e:
        logger.error(f"Error creating directory for file {file_path}: {e}")
        return ActionResult(error=f'Failed to create directory for {file_path}: {str(e)}', include_in_memory=True)



import asyncio
import os
import time
from playwright.async_api import Error as PlaywrightError
from playwright.async_api import Page, TimeoutError
from playwright.async_api import TimeoutError as PlaywrightTimeoutError
from app.helpers.tool_helpers import run_shell
from app.logger import logger
from app.tools.base import DEFAULT_WORKING_DIR
from app.types.browser_types import (
    BrowserClickAction, BrowserConsoleExecAction, BrowserConsoleViewAction, BrowserInputAction,
    BrowserMoveMouseAction, BrowserNavigateAction, BrowserPressKeyAction, BrowserRestartAction,
    BrowserScreenshotAction, BrowserScrollDownAction, BrowserScrollUpAction, BrowserSelectOptionAction,
    BrowserViewAction
)
from browser_use import ActionResult
from browser_use.browser.context import BrowserContext
from browser_helpers import HelperJs
from browser_manager import BrowserManager

__all__ = [
    'register_browser_actions'
]

TOTAL_TIMEOUT_MS = 45000
CONN_TIMEOUT_MS = 10000
default_headers = {
    'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.0.0 Safari/537.36',
    'Accept-Language': 'en-US,en;q=0.9'
}

def register_browser_actions(manager):
    # Define browser action handlers with lambda initial declarations
    action_browser_navigate = lambda params, browser: None  # 浏览器导航操作 - Browser navigation operation
    action_browser_click = lambda params, browser: None  # 浏览器点击操作 - Browser click operation
    action_browser_input = lambda params, browser: None  # 浏览器输入操作 - Browser input operation
    action_browser_view = lambda params, browser: None  # 浏览器查看操作 - Browser view operation
    action_browser_screenshot = lambda params, browser: None  # 浏览器截图操作 - Browser screenshot operation
    action_browser_scroll_down = lambda params, browser: None  # 浏览器向下滚动 - Browser scroll down
    action_browser_scroll_up = lambda params, browser: None  # 浏览器向上滚动 - Browser scroll up
    action_browser_press_key = lambda params, browser: None  # 浏览器按键操作 - Browser key press
    action_browser_select_option = lambda params, browser: None  # 浏览器选择选项 - Browser select option
    action_browser_console_exec = lambda params, browser: None  # 浏览器控制台执行 - Browser console execution
    action_browser_console_view = lambda params, browser: None  # 浏览器控制台查看 - Browser console view
    action_browser_restart = lambda params, browser: None  # 浏览器重启操作 - Browser restart
    action_browser_move_mouse = lambda params, browser: None  # 浏览器移动鼠标 - Browser mouse movement
    
    async def handle_navigation_timeout(page):
        """Handle navigation timeout by getting partial page details."""
        try:
            # Try to get current page details
            current_url = page.url
            title = await page.title()
            
            # Get the page height for scrolling calculations
            height = await page.evaluate("document.documentElement.scrollHeight")
            
            return ActionResult(
                url=current_url,
                title=title,
                result="Navigation timed out, but page partially loaded",
                pixels_below=max(0, height - await page.evaluate("window.innerHeight")),
                include_in_memory=True
            )
        except Exception as e:
            logger.error(f"Error handling navigation timeout: {e}")
            return ActionResult(
                error="Navigation timed out and failed to get partial page details",
                include_in_memory=True
            )
    
    async def handle_timeout(page):
        """Handle general timeout errors."""
        try:
            # Try to get current page details
            current_url = page.url
            title = await page.title()
            
            return ActionResult(
                url=current_url,
                title=title,
                result="Operation timed out",
                include_in_memory=True
            )
        except Exception as e:
            logger.error(f"Error handling timeout: {e}")
            return ActionResult(
                error="Operation timed out and failed to get page details",
                include_in_memory=True
            )
    
    browser_context = None

    async def get_browser_context(key):
        """Get or create browser context with specified key."""
        nonlocal browser_context
        if browser_context is None:
            browser_context = BrowserContext(manager, key)
            await browser_context.initialize()
        return browser_context
    
    async def get_page_details(browser, page, original_url, now):
        """Get page details including URL, title, and navigation status."""
        try:
            current_url = page.url
            title = await page.title()
            
            return {
                "url": current_url,
                "title": title,
                "navigation": current_url != original_url,
                "timestamp": now,
                "elapsed": time.time() - now
            }
        except Exception as e:
            logger.error(f"Error getting page details: {e}")
            return {
                "url": original_url,
                "title": "Unknown",
                "navigation": False,
                "timestamp": now,
                "elapsed": time.time() - now,
                "error": str(e)
            }
    
    # Redefine handlers with actual implementations
    
    async def action_browser_navigate_impl(params, browser):
        """Navigate browser to a URL."""
        page = browser.page
        url = params.url
        
        try:
            # Start navigation with a timeout
            async with asyncio.timeout(TOTAL_TIMEOUT_MS / 1000):
                logger.info(f"Navigating to URL: {url}")
                
                # Perform the navigation
                response = await page.goto(url, timeout=TOTAL_TIMEOUT_MS, wait_until="networkidle")
                
                # Wait for page to stabilize
                await asyncio.sleep(1)
                
                current_url = page.url
                title = await page.title()
                
                # Get the page height for scrolling calculations
                height = await page.evaluate("document.documentElement.scrollHeight")
                
                return ActionResult(
                    url=current_url,
                    title=title,
                    result=f"Navigated to {url}",
                    pixels_below=max(0, height - await page.evaluate("window.innerHeight")),
                    include_in_memory=True
                )
        except TimeoutError:
            return await handle_navigation_timeout(page)
        except Exception as e:
            logger.error(f"Navigation error: {e}")
            return ActionResult(error=f"Failed to navigate to {url}: {str(e)}", include_in_memory=True)
    
    async def action_browser_click_impl(params, browser):
        """Handle click actions on browser elements."""
        page = browser.page
        
        try:
            # Get page details before the action
            current_url = page.url
            title = await page.title()
            
            # If index is provided, click on element by index
            if params.index is not None:
                # Evaluate JavaScript to find clickable elements
                elements = await page.evaluate(HelperJs.FIND_CLICKABLE)
                
                if params.index >= len(elements):
                    return ActionResult(
                        error=f"Index {params.index} is out of range. Only {len(elements)} clickable elements found.",
                        include_in_memory=True
                    )
                
                # Get the element to click
                element = elements[params.index]
                
                # Calculate the center of the element
                x = element["rect"]["x"] + element["rect"]["width"] / 2
                y = element["rect"]["y"] + element["rect"]["height"] / 2
                
                # Scroll to the element
                await page.evaluate(f"window.scrollTo(0, {max(0, y - 100)})")
                
                # Click on the element
                await page.mouse.click(x, y)
            
            # If coordinates are provided, click at the specific location
            elif params.coordinate_x is not None and params.coordinate_y is not None:
                x = params.coordinate_x
                y = params.coordinate_y
                
                # Scroll to the coordinates
                await page.evaluate(f"window.scrollTo(0, {max(0, y - 100)})")
                
                # Click at the specified coordinates
                await page.mouse.click(x, y)
            
            else:
                return ActionResult(
                    error="Either index or coordinates must be provided for click action.",
                    include_in_memory=True
                )
            
            # Wait for any resulting navigation or page updates
            try:
                await page.wait_for_load_state("networkidle", timeout=5000)
            except:
                # Ignore timeout, page might not have navigated
                pass
            
            # Get updated page details
            new_url = page.url
            new_title = await page.title()
            
            # Determine if navigation occurred
            nav_result = "navigated to a new page" if new_url != current_url else "clicked successfully"
            
            # Get the page height for scrolling calculations
            height = await page.evaluate("document.documentElement.scrollHeight")
            
            return ActionResult(
                url=new_url,
                title=new_title,
                result=f"Click action {nav_result}",
                pixels_below=max(0, height - await page.evaluate("window.innerHeight")),
                include_in_memory=True
            )
        except Exception as e:
            logger.error(f"Click error: {e}")
            return ActionResult(error=f"Failed to perform click action: {str(e)}", include_in_memory=True)
    
    async def action_browser_input_impl(params, browser):
        """Handle input operations in browser."""
        page = browser.page
        
        try:
            # Get page details
            current_url = page.url
            title = await page.title()
            
            # If index is provided, type in element by index
            if params.index is not None:
                # Evaluate JavaScript to find input elements
                elements = await page.evaluate(HelperJs.FIND_INPUTS)
                
                if params.index >= len(elements):
                    return ActionResult(
                        error=f"Index {params.index} is out of range. Only {len(elements)} input elements found.",
                        include_in_memory=True
                    )
                
                # Get the element to input text into
                element = elements[params.index]
                
                # Calculate the center of the element
                x = element["rect"]["x"] + element["rect"]["width"] / 2
                y = element["rect"]["y"] + element["rect"]["height"] / 2
                
                # Scroll to the element
                await page.evaluate(f"window.scrollTo(0, {max(0, y - 100)})")
                
                # Click on the element to focus it
                await page.mouse.click(x, y)
                
                # Clear any existing text (select all and delete)
                await page.keyboard.press("Control+A")
                await page.keyboard.press("Delete")
                
                # Type the text
                await page.keyboard.type(params.text)
                
                # Press Enter if requested
                if params.press_enter:
                    await page.keyboard.press("Enter")
                    
                    # Wait for any resulting navigation
                    try:
                        await page.wait_for_load_state("networkidle", timeout=5000)
                    except:
                        # Ignore timeout, page might not have navigated
                        pass
            
            # If coordinates are provided, click at that location and input text
            elif params.coordinate_x is not None and params.coordinate_y is not None:
                x = params.coordinate_x
                y = params.coordinate_y
                
                # Scroll to the coordinates
                await page.evaluate(f"window.scrollTo(0, {max(0, y - 100)})")
                
                # Click at the specified coordinates to focus
                await page.mouse.click(x, y)
                
                # Clear any existing text (select all and delete)
                await page.keyboard.press("Control+A")
                await page.keyboard.press("Delete")
                
                # Type the text
                await page.keyboard.type(params.text)
                
                # Press Enter if requested
                if params.press_enter:
                    await page.keyboard.press("Enter")
                    
                    # Wait for any resulting navigation
                    try:
                        await page.wait_for_load_state("networkidle", timeout=5000)
                    except:
                        # Ignore timeout, page might not have navigated
                        pass
            
            else:
                return ActionResult(
                    error="Either index or coordinates must be provided for input action.",
                    include_in_memory=True
                )
            
            # Get updated page details
            new_url = page.url
            new_title = await page.title()
            
            # Determine if navigation occurred
            nav_result = "and navigated to a new page" if new_url != current_url else ""
            
            # Get the page height for scrolling calculations
            height = await page.evaluate("document.documentElement.scrollHeight")
            
            return ActionResult(
                url=new_url,
                title=new_title,
                result=f"Text input successful {nav_result}",
                pixels_below=max(0, height - await page.evaluate("window.innerHeight")),
                include_in_memory=True
            )
        except Exception as e:
            logger.error(f"Input error: {e}")
            return ActionResult(error=f"Failed to perform input action: {str(e)}", include_in_memory=True)
    
    async def action_browser_view_impl(params, browser):
        """View page content action."""
        page = browser.page
        
        try:
            # Reload the page if requested
            if params and params.reload:
                await page.reload(timeout=TOTAL_TIMEOUT_MS, wait_until="networkidle")
                await asyncio.sleep(1)
            
            # Get current page details
            current_url = page.url
            title = await page.title()
            
            # Extract page content using JavaScript
            content = await page.evaluate(HelperJs.EXTRACT_CONTENT)
            
            # Get the page height for scrolling calculations
            height = await page.evaluate("document.documentElement.scrollHeight")
            
            return ActionResult(
                url=current_url,
                title=title,
                result=content,
                pixels_below=max(0, height - await page.evaluate("window.innerHeight")),
                include_in_memory=True
            )
        except Exception as e:
            logger.error(f"View error: {e}")
            return ActionResult(error=f"Failed to view page content: {str(e)}", include_in_memory=True)
    
    async def action_browser_screenshot_impl(params, browser):
        """Take a screenshot of the current page."""
        page = browser.page
        
        try:
            # Reload the page if requested
            if params and params.reload:
                await page.reload(timeout=TOTAL_TIMEOUT_MS, wait_until="networkidle")
                await asyncio.sleep(1)
            
            # Get current page details
            current_url = page.url
            title = await page.title()
            
            # Define the file path for the screenshot
            file_path = params.file
            if not os.path.isabs(file_path):
                file_path = os.path.join(DEFAULT_WORKING_DIR, file_path)
            
            # Ensure directory exists
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
            
            # Take the screenshot
            await page.screenshot(path=file_path, full_page=True)
            
            # Get the page height for scrolling calculations
            height = await page.evaluate("document.documentElement.scrollHeight")
            
            return ActionResult(
                url=current_url,
                title=title,
                result=f"Screenshot saved to {file_path}",
                pixels_below=max(0, height - await page.evaluate("window.innerHeight")),
                include_in_memory=True
            )
        except Exception as e:
            logger.error(f"Screenshot error: {e}")
            return ActionResult(error=f"Failed to take screenshot: {str(e)}", include_in_memory=True)
    
    async def action_browser_scroll_down_impl(params, browser):
        """Scroll the page down."""
        page = browser.page
        
        try:
            # Get current page details
            current_url = page.url
            title = await page.title()
            
            # Get current scroll position and page height
            current_scroll = await page.evaluate("window.scrollY")
            window_height = await page.evaluate("window.innerHeight")
            page_height = await page.evaluate("document.documentElement.scrollHeight")
            
            # Calculate new scroll position
            if params and params.to_bottom:
                # Scroll to the bottom of the page
                new_scroll = page_height
            else:
                # Scroll down by one viewport height
                new_scroll = min(current_scroll + window_height, page_height - window_height)
            
            # Perform the scroll
            await page.evaluate(f"window.scrollTo(0, {new_scroll})")
            
            # Wait a moment for any lazy-loaded content
            await asyncio.sleep(0.5)
            
            # Recalculate page height (may have changed due to lazy loading)
            updated_page_height = await page.evaluate("document.documentElement.scrollHeight")
            updated_scroll = await page.evaluate("window.scrollY")
            
            # Calculate remaining pixels below viewport
            pixels_below = max(0, updated_page_height - (updated_scroll + window_height))
            
            scroll_result = "bottom of page" if params and params.to_bottom else f"position {updated_scroll}px"
            
            return ActionResult(
                url=current_url,
                title=title,
                result=f"Scrolled to {scroll_result}",
                pixels_below=pixels_below,
                include_in_memory=True
            )
        except Exception as e:
            logger.error(f"Scroll down error: {e}")
            return ActionResult(error=f"Failed to scroll down: {str(e)}", include_in_memory=True)
    
    async def action_browser_scroll_up_impl(params, browser):
        """Scroll the page up."""
        page = browser.page
        
        try:
            # Get current page details
            current_url = page.url
            title = await page.title()
            
            # Get current scroll position and page height
            current_scroll = await page.evaluate("window.scrollY")
            window_height = await page.evaluate("window.innerHeight")
            page_height = await page.evaluate("document.documentElement.scrollHeight")
            
            # Calculate new scroll position
            if params and params.to_top:
                # Scroll to the top of the page
                new_scroll = 0
            else:
                # Scroll up by one viewport height
                new_scroll = max(current_scroll - window_height, 0)
            
            # Perform the scroll
            await page.evaluate(f"window.scrollTo(0, {new_scroll})")
            
            # Wait a moment for any animations to complete
            await asyncio.sleep(0.5)
            
            # Get updated scroll position
            updated_scroll = await page.evaluate("window.scrollY")
            
            # Calculate remaining pixels below viewport
            pixels_below = max(0, page_height - (updated_scroll + window_height))
            
            scroll_result = "top of page" if params and params.to_top else f"position {updated_scroll}px"
            
            return ActionResult(
                url=current_url,
                title=title,
                result=f"Scrolled to {scroll_result}",
                pixels_below=pixels_below,
                include_in_memory=True
            )
        except Exception as e:
            logger.error(f"Scroll up error: {e}")
            return ActionResult(error=f"Failed to scroll up: {str(e)}", include_in_memory=True)
    
    async def action_browser_press_key_impl(params, browser):
        """Press a key in the browser."""
        page = browser.page
        
        try:
            # Get current page details
            current_url = page.url
            title = await page.title()
            
            # Press the specified key
            await page.keyboard.press(params.key)
            
            # Wait for any resulting navigation or page updates
            try:
                await page.wait_for_load_state("networkidle", timeout=5000)
            except:
                # Ignore timeout, page might not have navigated
                pass
            
            # Get updated page details
            new_url = page.url
            new_title = await page.title()
            
            # Determine if navigation occurred
            nav_result = "and navigated to a new page" if new_url != current_url else ""
            
            # Get the page height for scrolling calculations
            height = await page.evaluate("document.documentElement.scrollHeight")
            
            return ActionResult(
                url=new_url,
                title=new_title,
                result=f"Pressed key '{params.key}' {nav_result}",
                pixels_below=max(0, height - await page.evaluate("window.innerHeight")),
                include_in_memory=True
            )
        except Exception as e:
            logger.error(f"Key press error: {e}")
            return ActionResult(error=f"Failed to press key '{params.key}': {str(e)}", include_in_memory=True)
    
    async def action_browser_select_option_impl(params, browser):
        """Select an option from a dropdown menu."""
        page = browser.page
        
        try:
            # Get current page details
            current_url = page.url
            title = await page.title()
            
            # Find all select elements
            select_elements = await page.query_selector_all('select')
            
            if params.index >= len(select_elements):
                return ActionResult(
                    error=f"Index {params.index} is out of range. Only {len(select_elements)} select elements found.",
                    include_in_memory=True
                )
            
            # Get the select element
            select_element = select_elements[params.index]
            
            # Scroll to the element
            await select_element.scroll_into_view_if_needed()
            
            # Get all options in the select element
            options = await select_element.query_selector_all('option')
            
            if params.option >= len(options):
                return ActionResult(
                    error=f"Option index {params.option} is out of range. Only {len(options)} options found.",
                    include_in_memory=True
                )
            
            # Get the option value
            option_element = options[params.option]
            option_value = await option_element.get_attribute('value')
            option_text = await option_element.inner_text()
            
            # Select the option
            await select_element.select_option(value=option_value)
            
            # Wait for any resulting events or page updates
            await asyncio.sleep(0.5)
            
            # Get the page height for scrolling calculations
            height = await page.evaluate("document.documentElement.scrollHeight")
            
            return ActionResult(
                url=current_url,
                title=title,
                result=f"Selected option '{option_text}' from select element at index {params.index}",
                pixels_below=max(0, height - await page.evaluate("window.innerHeight")),
                include_in_memory=True
            )
        except Exception as e:
            logger.error(f"Select option error: {e}")
            return ActionResult(error=f"Failed to select option: {str(e)}", include_in_memory=True)
    
    async def action_browser_console_exec_impl(params, browser):
        """Execute JavaScript code in the browser console."""
        page = browser.page
        
        try:
            # Get current page details
            current_url = page.url
            title = await page.title()
            
            # Initialize console logging if needed
            await page.evaluate(HelperJs.INIT_CONSOLE_LOGGING)
            
            # Execute the JavaScript code
            result = await page.evaluate(params.javascript)
            
            # Convert result to string if it's not already
            if result is None:
                result_str = "undefined"
            elif isinstance(result, (dict, list)):
                result_str = str(result)
            else:
                result_str = str(result)
            
            # Get the page height for scrolling calculations
            height = await page.evaluate("document.documentElement.scrollHeight")
            
            return ActionResult(
                url=current_url,
                title=title,
                result=f"JavaScript executed successfully. Result: {result_str}",
                pixels_below=max(0, height - await page.evaluate("window.innerHeight")),
                include_in_memory=True
            )
        except Exception as e:
            logger.error(f"Console exec error: {e}")
            return ActionResult(error=f"Failed to execute JavaScript: {str(e)}", include_in_memory=True)
    
    async def action_browser_console_view_impl(params, browser):
        """View the console logs in the browser."""
        page = browser.page
        
        try:
            # Get current page details
            current_url = page.url
            title = await page.title()
            
            # Initialize console logging if needed
            init_result = await page.evaluate(HelperJs.INIT_CONSOLE_LOGGING)
            
            # Get console logs
            max_lines = params.max_lines if params and params.max_lines is not None else 100
            logs = await page.evaluate(HelperJs.CONSOLE_LOGS % max_lines)
            
            # Get the page height for scrolling calculations
            height = await page.evaluate("document.documentElement.scrollHeight")
            
            return ActionResult(
                url=current_url,
                title=title,
                result=f"Console logs:\n\n{logs}",
                pixels_below=max(0, height - await page.evaluate("window.innerHeight")),
                include_in_memory=True
            )
        except Exception as e:
            logger.error(f"Console view error: {e}")
            return ActionResult(error=f"Failed to view console logs: {str(e)}", include_in_memory=True)
    
    async def action_browser_restart_impl(params, browser):
        """Restart the browser instance."""
        try:
            # Close the current browser context and page
            await browser_context.close()
            
            # Recreate the browser context
            await browser_context.create()
            
            # Create a new page and navigate to the initial URL
            page = browser_context.page
            url = params.url
            
            # Setup default headers
            await page.set_extra_http_headers(default_headers)
            
            # Navigate to the specified URL
            response = await page.goto(url, timeout=TOTAL_TIMEOUT_MS, wait_until="networkidle")
            
            # Wait for page to stabilize
            await asyncio.sleep(1)
            
            current_url = page.url
            title = await page.title()
            
            # Get the page height for scrolling calculations
            height = await page.evaluate("document.documentElement.scrollHeight")
            
            return ActionResult(
                url=current_url,
                title=title,
                result=f"Browser restarted and navigated to {url}",
                pixels_below=max(0, height - await page.evaluate("window.innerHeight")),
                include_in_memory=True
            )
        except Exception as e:
            logger.error(f"Browser restart error: {e}")
            return ActionResult(error=f"Failed to restart browser: {str(e)}", include_in_memory=True)
    
    async def action_browser_move_mouse_impl(params, browser):
        """Move the mouse to specified coordinates."""
        page = browser.page
        
        try:
            # Get current page details
            current_url = page.url
            title = await page.title()
            
            # Extract coordinates
            x = params.coordinate_x
            y = params.coordinate_y
            
            # Scroll to ensure the target area is visible
            await page.evaluate(f"window.scrollTo(0, {max(0, y - 100)})")
            
            # Move the mouse to the specified coordinates
            await page.mouse.move(x, y)
            
            # Get the page height for scrolling calculations
            height = await page.evaluate("document.documentElement.scrollHeight")
            
            return ActionResult(
                url=current_url,
                title=title,
                result=f"Moved mouse to coordinates ({x}, {y})",
                pixels_below=max(0, height - await page.evaluate("window.innerHeight")),
                include_in_memory=True
            )
        except Exception as e:
            logger.error(f"Mouse movement error: {e}")
            return ActionResult(error=f"Failed to move mouse: {str(e)}", include_in_memory=True)
    
    # Assign the implementation functions to the lambda placeholders
    action_browser_navigate = action_browser_navigate_impl
    action_browser_click = action_browser_click_impl
    action_browser_input = action_browser_input_impl
    action_browser_view = action_browser_view_impl
    action_browser_screenshot = action_browser_screenshot_impl
    action_browser_scroll_down = action_browser_scroll_down_impl
    action_browser_scroll_up = action_browser_scroll_up_impl
    action_browser_press_key = action_browser_press_key_impl
    action_browser_select_option = action_browser_select_option_impl
    action_browser_console_exec = action_browser_console_exec_impl
    action_browser_console_view = action_browser_console_view_impl
    action_browser_restart = action_browser_restart_impl
    action_browser_move_mouse = action_browser_move_mouse_impl
    
    # Create browser context
    browser_context = BrowserContext(manager)
    
    # Register action handlers with the manager
    manager.register_action_handler("browser_navigate", action_browser_navigate)
    manager.register_action_handler("browser_click", action_browser_click)
    manager.register_action_handler("browser_input", action_browser_input)
    manager.register_action_handler("browser_view", action_browser_view)
    manager.register_action_handler("browser_screenshot", action_browser_screenshot)
    manager.register_action_handler("browser_scroll_down", action_browser_scroll_down)
    manager.register_action_handler("browser_scroll_up", action_browser_scroll_up)
    manager.register_action_handler("browser_press_key", action_browser_press_key)
    manager.register_action_handler("browser_select_option", action_browser_select_option)
    manager.register_action_handler("browser_console_exec", action_browser_console_exec)
    manager.register_action_handler("browser_console_view", action_browser_console_view)
    manager.register_action_handler("browser_restart", action_browser_restart)
    manager.register_action_handler("browser_move_mouse", action_browser_move_mouse)
    
    return browser_context


from dataclasses import dataclass, field
from functools import cached_property
from typing import TYPE_CHECKING, Dict, List, Optional
from browser_use.dom.history_tree_processor.view import CoordinateSet, HashedDomElement, ViewportInfo

# Avoid circular import issues
if TYPE_CHECKING:
    from .views import DOMElementNode


@dataclass(frozen=False)
class DOMBaseNode:
    is_visible: bool
    # Use None as default and set parent later to avoid circular reference issues
    parent: Optional['DOMElementNode'] = None


@dataclass(frozen=False)
class DOMTextNode(DOMBaseNode):
    # Fix: Use field() with default_factory to avoid non-default argument following default argument
    text: str = field(default="")  # Give text a default value
    type: str = 'TEXT_NODE'

    def has_parent_with_highlight_index(self) -> bool:
        current = self.parent
        while current is not None:
            # stop if the element has a highlight index (will be handled separately)
            if current.highlight_index is not None:
                return True

            current = current.parent
        return False

    def is_parent_in_viewport(self) -> bool:
        if self.parent is None:
            return False
        return self.parent.is_in_viewport

    def is_parent_top_element(self) -> bool:
        if self.parent is None:
            return False
        return self.parent.is_top_element


@dataclass(frozen=False)
class DOMElementNode(DOMBaseNode):
    """
    xpath: the xpath of the element from the last root node (shadow root or iframe OR document if no shadow root or iframe).
    To properly reference the element we need to recursively switch the root node until we find the element (work you way up the tree with `.parent`)
    """

    # Fix: Give default values to all required parameters that follow a parameter with a default
    tag_name: str = field(default="")
    xpath: str = field(default="")
    attributes: Dict[str, str] = field(default_factory=dict)
    children: List[DOMBaseNode] = field(default_factory=list)
    desc: str = field(default="")
    is_interactive: bool = False
    is_scroll_element: bool = False
    is_top_element: bool = False
    is_in_viewport: bool = False
    shadow_root: bool = False
    highlight_index: Optional[int] = None
    viewport_coordinates: Optional[CoordinateSet] = None
    page_coordinates: Optional[CoordinateSet] = None
    viewport_info: Optional[ViewportInfo] = None

    def __repr__(self) -> str:
        tag_str = f'<{self.tag_name}'

        # Add attributes
        for key, value in self.attributes.items():
            tag_str += f' {key}="{value}"'
        tag_str += '>'

        # Add extra info
        extras = []
        if self.is_interactive:
            extras.append('interactive')
        if self.is_top_element:
            extras.append('top')
        if self.shadow_root:
            extras.append('shadow-root')
        if self.highlight_index is not None:
            extras.append(f'highlight:{self.highlight_index}')
        if self.is_in_viewport:
            extras.append('in-viewport')

        if extras:
            tag_str += f' [{", ".join(extras)}]'

        return tag_str

    @cached_property
    def hash(self) -> HashedDomElement:
        from browser_use.dom.history_tree_processor.service import (
            HistoryTreeProcessor,
        )

        return HistoryTreeProcessor._hash_dom_element(self)

    def get_all_text_till_next_clickable_element(self, max_depth: int = -1) -> str:
        text_parts = []

        def collect_text(node: DOMBaseNode, current_depth: int) -> None:
            if max_depth != -1 and current_depth > max_depth:
                return

            # Skip this branch if we hit a highlighted element (except for the current node)
            if isinstance(node, DOMElementNode) and node != self and node.highlight_index is not None:
                return

            if isinstance(node, DOMTextNode):
                text_parts.append(node.text)
            elif isinstance(node, DOMElementNode):
                for child in node.children:
                    collect_text(child, current_depth + 1)

        collect_text(self, 0)
        return '\n'.join(text_parts).strip()

    # Note: the time_execution_sync decorator is missing from the provided code
    # Commenting it out for now
    # @time_execution_sync('--clickable_elements_to_string')
    def clickable_elements_to_string(self, include_attributes: list[str] | None = None) -> str:
        """Convert the processed DOM content to HTML."""
        formatted_text = []

        def process_node(node: DOMBaseNode, depth: int) -> None:
            if isinstance(node, DOMElementNode):
                # Add element with highlight_index
                if node.highlight_index is not None:
                    attributes_str = ''
                    if(node.is_scroll_element):
                        text = ''
                    else:
                        text = node.get_all_text_till_next_clickable_element()
                    if include_attributes:
                        attributes = list(
                            set(
                                [
                                    str(value)
                                    for key, value in node.attributes.items()
                                    if key in include_attributes and value != node.tag_name
                                ]
                            )
                        )
                        if text in attributes:
                            attributes.remove(text)
                        attributes_str = ';'.join(attributes)
                    if node.desc:
                        line = f'[{node.highlight_index}]:({node.desc})<{node.tag_name}'
                        for key in node.attributes.keys():
                            if key != 'style':
                                line += f' {key}="{node.attributes[key]}"'
                        line += ' '
                    else:
                        line = f'[{node.highlight_index}]:<{node.tag_name}'
                        for key in node.attributes.keys():
                            if key != 'style':
                                line += f' {key}="{node.attributes[key]}"'
                    if attributes_str:
                        line += f'{attributes_str}'
                    if text:
                        if attributes_str:
                            line += f'>{text}'
                        else:
                            line += f'{text}'

                    line += '/>'
                    
                    formatted_text.append(line)

                # Process children regardless
                for child in node.children:
                    process_node(child, depth + 1)

            elif isinstance(node, DOMTextNode):
                # Add text only if it doesn't have a highlighted parent
                if not node.has_parent_with_highlight_index() and node.is_visible:
                    formatted_text.append(f'{node.text}')

        process_node(self, 0)
        return '\n'.join(formatted_text)

    def get_file_upload_element(self, check_siblings: bool = True) -> Optional['DOMElementNode']:
        # Check if current element is a file input
        if self.tag_name == 'input' and self.attributes.get('type') == 'file':
            return self

        # Check children
        for child in self.children:
            if isinstance(child, DOMElementNode):
                result = child.get_file_upload_element(check_siblings=False)
                if result:
                    return result

        # Check siblings only for the initial call
        if check_siblings and self.parent:
            for sibling in self.parent.children:
                if sibling is not self and isinstance(sibling, DOMElementNode):
                    result = sibling.get_file_upload_element(check_siblings=False)
                    if result:
                        return result

        return None

    def get_advanced_css_selector(self) -> str:
        from browser_use.browser.context import BrowserContext
        return BrowserContext._enhanced_css_selector_for_element(self)


class ElementTreeSerializer:
    @staticmethod
    def serialize_clickable_elements(element_tree):
        return element_tree.clickable_elements_to_string()
    
    @staticmethod
    def dom_element_node_to_json(element_tree):
        def process_node(node):
            if isinstance(node, DOMTextNode):
                return {
                    'type': 'text',
                    'text': node.text
                }
            elif isinstance(node, DOMElementNode):
                return {
                    'type': 'element',
                    'tag_name': node.tag_name,
                    'attributes': node.attributes,
                    'highlight_index': node.highlight_index,
                    'children': [process_node(child) for child in node.children]
                }
            return None
            
        return process_node(element_tree)


SelectorMap = Dict[int, DOMElementNode]


@dataclass
class DOMState:
    element_tree: DOMElementNode
    selector_map: SelectorMap


#!/usr/bin/env python3
"""
示例：如何通过API调用使用Orion服务

此示例展示：
1. 如何连接到运行中的Orion服务
2. 如何调用浏览器API
3. 如何调用终端API
4. 如何调用文本编辑器API
"""

import asyncio
import json
import httpx
from websockets.client import connect

API_BASE = "http://localhost:8330"  # Orion服务器基础URL

async def browser_api_example():
    """浏览器API使用示例"""
    print("=== 浏览器API示例 ===")
    
    async with httpx.AsyncClient() as client:
        # 获取浏览器状态
        response = await client.get(f"{API_BASE}/browser/status")
        status = response.json()
        print(f"浏览器状态: {'健康' if status.get('healthy') else '不健康'}")
        
        # 执行浏览器操作（导航到网页）
        browser_action = {
            "action": "navigate",
            "args": {"url": "https://www.baidu.com"}
        }
        response = await client.post(f"{API_BASE}/browser/action", json=browser_action)
        result = response.json()
        
        if result["status"] == "success":
            print(f"导航成功 - 标题: {result['result']['title']}")
            print(f"当前URL: {result['result']['url']}")
        else:
            print(f"导航失败: {result['error']}")

async def terminal_api_example():
    """终端API使用示例"""
    print("\n=== 终端API示例 ===")
    
    # 创建一个终端ID
    terminal_id = "api_example"
    
    # 使用WebSocket连接到终端
    async with connect(f"ws://localhost:8330/terminal") as websocket:
        # 发送初始化消息
        await websocket.send(json.dumps({
            "type": "init",
            "terminal_id": terminal_id
        }))
        
        # 接收初始化响应
        response = json.loads(await websocket.recv())
        print(f"终端初始化: {response.get('success', False)}")
        
        # 发送命令
        await websocket.send(json.dumps({
            "type": "input",
            "content": "echo 'Hello from API'; ls -la\n"
        }))
        
        # 接收输出 (最多接收5条消息或等待5秒)
        count = 0
        try:
            while count < 5:
                response = json.loads(await asyncio.wait_for(websocket.recv(), timeout=5))
                print(f"终端输出: {response.get('content', '')}")
                count += 1
        except asyncio.TimeoutError:
            print("终端输出接收超时")
    
    # 使用HTTP API获取终端历史
    async with httpx.AsyncClient() as client:
        response = await client.get(f"{API_BASE}/terminal/{terminal_id}/view?full=true")
        result = response.json()
        
        if result["status"] == "success":
            print(f"\n获取终端历史成功, 共 {len(result['output'])} 行")
        else:
            print(f"获取终端历史失败: {result.get('error', 'unknown error')}")
        
        # 重置终端
        response = await client.post(f"{API_BASE}/terminal/{terminal_id}/reset")
        result = response.json()
        print(f"终端重置: {'成功' if result['status'] == 'success' else '失败'}")

async def text_editor_api_example():
    """文本编辑器API使用示例"""
    print("\n=== 文本编辑器API示例 ===")
    
    async with httpx.AsyncClient() as client:
        # 创建测试文件
        test_file = "/tmp/orion_api_test.txt"
        
        # 写入文件
        action = {
            "action": "write_file",
            "target_file": test_file,
            "content": "这是通过API创建的测试文件\n第二行内容\n第三行内容"
        }
        
        response = await client.post(f"{API_BASE}/text_editor", json=action)
        result = response.json()
        print(f"文件写入: {'成功' if result['status'] == 'success' else '失败'}")
        
        # 读取文件
        action = {
            "action": "read_file",
            "target_file": test_file
        }
        
        response = await client.post(f"{API_BASE}/text_editor", json=action)
        result = response.json()
        
        if result["status"] == "success":
            print(f"文件内容:\n{result['result']}")
        else:
            print(f"读取文件失败: {result.get('error', 'unknown error')}")
        
        # 删除文件
        action = {
            "action": "delete_file",
            "target_file": test_file
        }
        
        response = await client.post(f"{API_BASE}/text_editor", json=action)
        result = response.json()
        print(f"删除文件: {'成功' if result['status'] == 'success' else '失败'}")

async def main():
    """运行所有API示例"""
    print("Orion API调用示例\n")
    print("注意: 请确保Orion服务已在localhost:8330上运行\n")
    
    try:
        # 运行浏览器API示例
        await browser_api_example()
        
        # 运行终端API示例
        await terminal_api_example()
        
        # 运行文本编辑器API示例
        await text_editor_api_example()
    except httpx.ConnectError:
        print("\n错误: 无法连接到Orion服务器。请确保服务器正在运行。")
    except Exception as e:
        print(f"\n错误: {e}")

if __name__ == "__main__":
    asyncio.run(main()) 


import base64
import pytest
from browser_use.browser.browser import Browser, BrowserConfig

@pytest.fixture
def browser():
    """Create a browser instance for testing"""
    browser_config = BrowserConfig(headless=False)
    browser_instance = Browser(config=browser_config)
    yield browser_instance
    # Cleanup after tests
    # This would normally include code to close the browser

def test_take_full_page_screenshot(browser):
    """Test taking a full page screenshot"""
    browser.go_to_url('https://example.com')
    screenshot_data = browser.take_screenshot(full_page=True)
    
    # Verify the screenshot data is valid base64
    assert screenshot_data is not None
    try:
        base64.b64decode(screenshot_data)
        assert True
    except Exception:
        assert False, "Screenshot data is not valid base64"

if __name__ == '__main__':
    test_take_full_page_screenshot(Browser(config=BrowserConfig(headless=False)))


import asyncio
import json
import math
import mimetypes
import os
import sys
import time
from pathlib import Path
from typing import Dict
import aiohttp
import httpx
from fastapi import Body, FastAPI, HTTPException, Query, WebSocket
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import FileResponse
from pydantic import BaseModel
from app.logger import logger
from app.models import DownloadRequest, FileUploadRequest, MultipartUploadRequest, MultipartUploadResponse, ZipAndUploadRequest, ZipAndUploadResponse
from app.router import TimedRoute
from app.tools.terminal.terminal_socket import terminal_socket
from app.tools.terminal.terminal_manager import terminal_manager
from app.tools.base import get_file_path, validate_dir_path, validate_file_path
from app.tools.browser.browser_manager import BrowserDeadError, BrowserManager, PageDeadError
from app.types.messages import BrowserActionRequest, BrowserActionResponse, TerminalApiResponse, TerminalWriteApiRequest, TextEditorCreateRequest, TextEditorDeleteRequest, TextEditorDirRequest, TextEditorFindContentRequest, TextEditorFindFileRequest, TextEditorMoveRequest, TextEditorStrReplaceRequest, TextEditorViewRequest, TextEditorWriteRequest
import app.tools.editor as editor
import app.tools.file as file
import app.tools.zip as zip

app = FastAPI()
app.router.route_class = TimedRoute
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

MULTIPART_THRESHOLD = 10485760  # 10MB

@app.post("/file/upload")
@file.handle_error
async def upload_file(request: FileUploadRequest = Body(...)):
    """Upload a file to presigned_url. If file size exceeds threshold, return size information instead."""
    file_path = validate_file_path(get_file_path(request.file_path))
    file_size = file_path.stat().st_size
    content_type = mimetypes.guess_type(str(file_path))[0] or "application/octet-stream"
    file_name = file_path.name
    
    if file_size > MULTIPART_THRESHOLD:
        return {
            "status": "need_multipart",
            "message": "File size exceeds single upload limit",
            "file_name": file_name,
            "content_type": content_type,
            "file_size": file_size,
            "need_multipart": True,
            "recommended_part_size": MULTIPART_THRESHOLD,
            "estimated_parts": file_size // MULTIPART_THRESHOLD + 1
        }
    
    with open(file_path, 'rb') as f:
        content = f.read()
    
    upload_result = await file.upload_to_presigned_url(
        data=content,
        presigned_url=request.presigned_url
    )
    
    if not upload_result:
        raise Exception("File uploaded failed")
    
    if not upload_result['success']:
        raise Exception(f"File uploaded failed: {upload_result['error']}")
    
    return {
        "status": "success",
        "message": "File uploaded successfully",
        "file_name": file_name,
        "content_type": content_type,
        "file_size": file_size,
        "need_multipart": False,
        "upload_result": {"success": True, "uploaded": True}
    }

@app.post("/file/multipart_upload")
@file.handle_error
async def multipart_upload_file(request: MultipartUploadRequest = Body(...)):
    """Upload file chunks using presigned URLs"""
    file_path = validate_file_path(get_file_path(request.file_path))
    file_size = file_path.stat().st_size
    expected_parts = math.ceil(file_size / request.part_size)
    
    if len(request.presigned_urls) != expected_parts:
        raise Exception(f"Number of presigned URLs ({len(request.presigned_urls)}) does not match expected parts ({expected_parts})")
    
    results = await file.upload_parts_to_presigned_url(str(file_path), request.presigned_urls, request.part_size, request.max_concurrent)
    
    successful_parts = [r.part_number for r in results if r.success]
    failed_parts = [r.part_number for r in results if not r.success]
    
    response = MultipartUploadResponse(
        status="success" if not failed_parts else "partial_success",
        message="All parts uploaded successfully" if not failed_parts else f"Uploaded {len(successful_parts)}/{len(results)} parts successfully",
        file_name=file_path.name,
        parts_results=results,
        successful_parts=successful_parts,
        failed_parts=failed_parts
    )
    
    return response

@app.post("/file/download")
@file.handle_error
async def download_files(request: DownloadRequest = Body(...)):
    """Download files into vm"""
    results = []
    
    async with aiohttp.ClientSession() as session:
        tasks = [file.download_file(session, request.folder_path, item) for item in request.files]
        results = await asyncio.gather(*tasks)
    
    success_count = sum(1 for r in results if r.success)
    fail_count = len(results) - success_count
    
    return {
        "status": "completed",
        "total": len(results),
        "success_count": success_count,
        "fail_count": fail_count,
        "results": results
    }

@app.get("/file")
@file.handle_error
async def get_file(path: str):
    """Download a file from vm"""
    file_path = validate_file_path(get_file_path(path))
    
    return FileResponse(
        path=str(file_path),
        filename=file_path.name,
        media_type="application/octet-stream"
    )

@app.post("/dir/upload")
@zip.handle_error
async def dir_upload(request: ZipAndUploadRequest):
    """Zip a directory (excluding node_modules) and upload"""
    directory_path = validate_dir_path(get_file_path(request.directory))
    project_name = os.path.basename(directory_path)
    output_zip = f"/tmp/{project_name}.zip"
    zip_result = zip.zip_dir(directory_path, output_zip)
    
    if not zip_result['success']:
        return ZipAndUploadResponse(
            status="error",
            message="Failed to create zip file",
            error=zip_result['error']
        )
    
    if not os.path.exists(output_zip):
        return ZipAndUploadResponse(
            status="error",
            message="Zip file was not created",
            error="Zip operation failed"
        )
    
    with open(output_zip, 'rb') as f:
        content = f.read()
    
    upload_result = await file.upload_to_presigned_url(
        data=content,
        presigned_url=request.presigned_url
    )
    
    if not upload_result:
        raise Exception("Zip file uploaded failed")
    
    if not upload_result['success']:
        raise Exception(f"Zip file uploaded failed: {upload_result['error']}")
    
    os.remove(output_zip)
    
    return ZipAndUploadResponse(
        status="success",
        message=f"Successfully processed {request.directory} and uploaded to presigned_url"
    )

# Initialize browser manager
browser_manager = BrowserManager(headless=False)

@app.get("/browser/status")
async def browser_status():
    """Endpoint for browser status"""
    try:
        tabs = await browser_manager.health_check()
        return {"healthy": True, "tabs": tabs}
    except BrowserDeadError as e:
        logger.error(f"Browser Error: {e}")
        return {"healthy": False, "tabs": []}

@app.post("/browser/action")
async def browser_action(cmd: BrowserActionRequest = Body()):
    """Endpoint for browser action"""
    async def execute_with_retry():
        timeout = 60
        try:
            return await asyncio.wait_for(
                browser_manager.execute_action(cmd),
                timeout=timeout
            )
        except asyncio.TimeoutError:
            error_msg = f"Browser action timed out after {timeout}s, new tab created and opened target:blank."
            logger.error(error_msg)
            await browser_manager.recreate_page()
            raise PageDeadError(error_msg)
    
    try:
        logger.info(f"start handling browser action {repr(cmd)}")
        result = await execute_with_retry()
        
        logger.info("\n".join([
            "Browser action result:",
            "title: " + result.title,
            "url: " + result.url,
            "result: " + result.result
        ]))
        
        return BrowserActionResponse(
            status="success",
            result=result,
            error=None
        ).model_dump()
    except PageDeadError as e:
        await browser_manager.recreate_page()
        logger.error(e)
        return BrowserActionResponse(
            status="error",
            result=None,
            error=str(e)
        ).model_dump()
    except Exception as e:
        logger.error(f"Browser Error: {e}")
        return BrowserActionResponse(
            status="error",
            result=None,
            error=str(e)
        ).model_dump()

class RecordingUpdateRequest(BaseModel):
    data: Dict

def _get_recording_path():
    """获取 recording_task.json 文件路径，按照 service 的逻辑"""
    if sys.platform.startswith('linux'):
        cache_dir = '/home/ubuntu/workspace/.cache'
    elif sys.platform == 'darwin':
        cache_dir = os.path.join(Path.home(), 'Library', 'Caches')
    elif sys.platform == 'win32':
        cache_dir = os.environ.get('LOCALAPPDATA', os.path.join(Path.home(), 'AppData', 'Local'))
    else:
        raise RuntimeError(f'Unsupported platform: {sys.platform}')
    
    recording_dir = os.path.join(cache_dir, 'orion-recording')
    os.makedirs(recording_dir, exist_ok=True)
    
    return os.path.join(recording_dir, 'recording_task.json')

@app.post("/browser/recording/update")
async def update_recording(request: RecordingUpdateRequest = Body(...)):
    """更新 recording_task.json 文件
    
    根据 service 的路径逻辑更新 JSON 文件，如果文件不存在则创建
    
    Args:
        request: RecordingUpdateRequest containing JSON data
        
    Returns:
        Dict with status and file info
        
    Raises:
        HTTPException: If file operations fail
    """
    try:
        recording_path = _get_recording_path()
        logger.info(f"更新录制文件: {recording_path}")
        
        # 写入 JSON 数据
        with open(recording_path, 'w', encoding='utf-8') as f:
            json.dump(request.data, f, ensure_ascii=False, indent=2)
        
        return {
            'status': 'success',
            'message': 'Recording data updated successfully',
            'file_path': recording_path
        }
    except Exception as e:
        logger.error(f"Error updating recording file: {e}")
        raise HTTPException(status_code=500, detail=f"Failed to update recording file: {str(e)}")

@app.post("/text_editor/dir")
@editor.handle_error
async def text_editor_dir(cmd: TextEditorDirRequest):
    return await editor.dir(cmd.path)

@app.post("/text_editor/view")
@editor.handle_error
async def text_editor_view(cmd: TextEditorViewRequest):
    return await editor.view(cmd.path, cmd.view_range, cmd.sudo)

@app.post("/text_editor/create")
@editor.handle_error
async def text_editor_create(cmd: TextEditorCreateRequest):
    return await editor.create(cmd.path, cmd.file_text, cmd.sudo)

@app.post("/text_editor/write")
@editor.handle_error
async def text_editor_write(cmd: TextEditorWriteRequest):
    return await editor.write(cmd.path, cmd.file_text, cmd.sudo, cmd.append, cmd.trailing_newline, cmd.leading_newline)

@app.post("/text_editor/replace")
@editor.handle_error
async def text_editor_replace(cmd: TextEditorStrReplaceRequest):
    return await editor.replace(cmd.path, cmd.old_str, cmd.new_str, cmd.sudo)

@app.post("/text_editor/find_content")
@editor.handle_error
async def text_editor_find_content(cmd: TextEditorFindContentRequest):
    return await editor.find_content(cmd.path, cmd.regex, cmd.sudo)

@app.post("/text_editor/find_file")
@editor.handle_error
async def text_editor_find_file(cmd: TextEditorFindFileRequest):
    return await editor.find_file(cmd.path, cmd.glob)

@app.post("/text_editor/move")
@editor.handle_error
async def text_editor_move(cmd: TextEditorMoveRequest):
    return await editor.move(cmd.path, cmd.new_path, cmd.sudo)

@app.post("/text_editor/delete")
@editor.handle_error
async def text_editor_delete(cmd: TextEditorDeleteRequest):
    return await editor.delete(cmd.path, cmd.sudo)

@app.websocket("/terminal")
async def websocket_endpoint(ws: WebSocket):
    await terminal_socket.handle_connection(ws)

@app.post("/terminal/{terminal_id}/reset")
async def reset_terminal(terminal_id: str):
    try:
        terminal = await terminal_manager.get_terminal(terminal_id)
        await terminal.reset()
        return TerminalApiResponse(
            status="success",
            result="terminal reset success",
            terminal_id=terminal_id,
            output=[]
        )
    except Exception as e:
        logger.error(f"Error resetting terminal: {e}")
        raise HTTPException(status_code=500, detail=str(e))

@app.post("/terminal/reset-all")
async def reset_all_terminals():
    try:
        for terminal in terminal_manager.terminals.values():
            await terminal.reset()
        
        return TerminalApiResponse(
            status="success",
            result="all terminals reset success",
            terminal_id="",
            output=[]
        )
    except Exception as e:
        logger.error(f"Error resetting all terminals: {e}")
        raise HTTPException(status_code=500, detail=str(e))

@app.get("/terminal/{terminal_id}/view")
async def view_terminal(terminal_id: str, full: bool = Query(True)):
    try:
        terminal = await terminal_manager.get_terminal(terminal_id)
        history = terminal.get_history(True, full)
        
        return TerminalApiResponse(
            status="success",
            result="terminal view success",
            terminal_id=terminal_id,
            output=history
        )
    except Exception as e:
        logger.error(f"Error viewing terminal: {e}")
        raise HTTPException(status_code=500, detail=str(e))

@app.post("/terminal/{terminal_id}/kill")
async def kill_terminal_process(terminal_id: str):
    try:
        terminal = await terminal_manager.get_terminal(terminal_id)
        await terminal.kill_process()
        
        history = terminal.get_history(True, False)
        
        return TerminalApiResponse(
            status="success",
            result="terminal process killed",
            terminal_id=terminal_id,
            output=history
        )
    except Exception as e:
        logger.error(f"Error killing terminal process: {e}")
        raise HTTPException(status_code=500, detail=str(e))

@app.post("/terminal/{terminal_id}/write")
async def write_terminal_process(terminal_id: str, cmd: TerminalWriteApiRequest):
    try:
        terminal = await terminal_manager.get_terminal(terminal_id)
        await terminal.write_to_process(cmd.text, cmd.enter if cmd.enter is not None else False)
        
        # Allow time for the process to respond
        await asyncio.sleep(1)
        
        history = terminal.get_history(True, False)
        
        return TerminalApiResponse(
            status="success",
            result="write terminal process success",
            terminal_id=terminal_id,
            output=history
        )
    except Exception as e:
        logger.error(f"Error killing terminal process: {e}")
        raise HTTPException(status_code=500, detail=str(e))

class InitSandboxRequest(BaseModel):
    secrets: Dict[str, str]

@app.post("/init-sandbox")
async def init_sandbox(request: InitSandboxRequest):
    """初始化沙箱环境  # Initialize sandbox environment

    接收 secrets 并写入到用户的 .secrets 目录下，每个 secret 作为单独的文件  # Receive secrets and write them to the user's .secrets directory, each secret as a separate file
    - secrets 目录会在 $HOME/.secrets 下创建  # The secrets directory will be created under $HOME/.secrets
    - 每个 secret 的 key 作为文件名  # Each secret's key is used as the filename
    - 如果文件已存在且内容不同，会将原文件备份（添加时间戳后缀）  # If the file already exists with different content, the original file will be backed up (with a timestamp suffix)

    Args:
        request: InitSandboxRequest containing secrets dictionary

    Returns:
        Dict with status and processed files info

    Raises:
        HTTPException: If HOME environment variable is not set or other errors
    """
    try:
        home_dir = os.getenv('WORKDIR')
        if not home_dir:
            raise HTTPException(status_code=500, detail="HOME environment variable is not set")
            
        secrets_dir = os.path.join(home_dir, '.secrets')
        
        # Create secrets directory if it doesn't exist
        os.makedirs(secrets_dir, exist_ok=True)
        os.chmod(secrets_dir, 0o700)  # rwx------
        
        processed_files = []
        
        for key, value in request.secrets.items():
            secret_file = os.path.join(secrets_dir, key)
            
            if os.path.exists(secret_file):
                try:
                    with open(secret_file, 'r') as f:
                        current_content = f.read()
                    
                    if current_content == value:
                        processed_files.append({
                            'key': key,
                            'action': 'skipped',
                            'reason': 'content unchanged'
                        })
                        continue
                    
                    if current_content != value:
                        # Backup the existing file with timestamp
                        timestamp = time.strftime('%Y%m%d_%H%M%S')
                        backup_file = f"{secret_file}.{timestamp}"
                        os.rename(secret_file, backup_file)
                        processed_files.append({
                            'key': key,
                            'action': 'backed_up',
                            'backup_file': backup_file
                        })
                except Exception as e:
                    logger.error(f"Error reading existing secret file {key}: {e}")
                    raise HTTPException(status_code=500, detail=f"Failed to process existing secret file {key}: {str(e)}")
            
            try:
                with open(secret_file, 'w') as f:
                    f.write(value)
                
                os.chmod(secret_file, 0o600)  # rw-------
                
                processed_files.append({
                    'key': key,
                    'action': 'updated' if os.path.exists(secret_file) else 'created'
                })
            except Exception as e:
                logger.error(f"Error writing secret file {key}: {e}")
                raise HTTPException(status_code=500, detail=f"Failed to write secret file {key}: {str(e)}")
        
        return {
            'status': 'ok',
            'secrets_dir': secrets_dir,
            'processed_files': processed_files
        }
    except Exception as e:
        logger.error(f"Error processing secrets: {e}")
        raise HTTPException(status_code=500, detail=f"Failed to process secrets: {str(e)}")

@app.get("/healthz")
async def healthz():
    """Health check endpoint."""
    # If browser is set to start automatically, create the task but don't await it
    if browser_manager.status == "started":
        asyncio.create_task(browser_manager.initialize())
    
    return {"status": "ok"}


from __future__ import annotations
import json
import logging
from datetime import datetime
from typing import Dict, List, Optional, Type
from langchain_anthropic import ChatAnthropic
from langchain_core.language_models import BaseChatModel
from langchain_core.messages import AIMessage, BaseMessage, HumanMessage, SystemMessage, ToolMessage
from langchain_openai import ChatOpenAI
from browser_use.agent.message_manager.views import MessageHistory, MessageMetadata
from browser_use.agent.prompts import AgentMessagePrompt, SystemPrompt
from browser_use.agent.views import ActionResult, AgentOutput, AgentStepInfo, MessageManagerState
from browser_use.browser.views import BrowserState
from pydantic import BaseModel

logger = logging.getLogger(__name__)


class MessageManagerSettings(BaseModel):
    max_input_tokens: int = 128000
    estimated_characters_per_token: int = 3
    image_tokens: int = 800
    include_attributes: List[str] = []
    message_context: Optional[str] = None
    sensitive_data: Optional[Dict[str, str]] = None
    available_file_paths: Optional[List[str]] = None


class MessageManager:
    def __init__(
        self,
        llm: Optional[BaseChatModel] = None,
        task: str = "",
        action_descriptions: Optional[List[str]] = None,
        system_prompt_class: Optional[Type[SystemPrompt]] = None,
        max_input_tokens: int = 128000,
        estimated_characters_per_token: int = 3,
        image_tokens: int = 800,
        include_attributes: Optional[List[str]] = None,
        max_error_length: int = 100,
        max_actions_per_step: int = 5,
        message_context: Optional[str] = None,
        sensitive_data: Optional[Dict[str, str]] = None,
        system_message: Optional[SystemMessage] = None,
        settings: Optional[MessageManagerSettings] = None,
        state: Optional[MessageManagerState] = None,
    ):
        # Support both old and new initialization patterns
        if settings is not None:
            # New style initialization
            self.task = task
            self.settings = settings
            self.state = state or MessageManagerState()
            self.system_prompt = system_message
            # Only initialize messages if state is empty
            if len(self.state.history.messages) == 0:
                self._init_messages()
        else:
            # Old style initialization
            self.llm = llm
            self.system_prompt_class = system_prompt_class
            self.max_input_tokens = max_input_tokens
            self.history = MessageHistory()
            self.task = task
            self.action_descriptions = action_descriptions or []
            self.estimated_characters_per_token = estimated_characters_per_token
            self.IMG_TOKENS = image_tokens
            self.include_attributes = include_attributes or []
            self.max_error_length = max_error_length
            self.message_context = message_context
            self.sensitive_data = sensitive_data
            self.tool_id = 1

            # Initialize with system prompt
            system_message = self.system_prompt_class(
                self.action_descriptions,
                max_actions_per_step=max_actions_per_step
            ).get_system_message()
            self._add_message_with_tokens(system_message)
            self.system_prompt = system_message

            # Add context message if provided
            if self.message_context:
                context_message = HumanMessage(content='Context for the task' + self.message_context)
                self._add_message_with_tokens(context_message)

            # Add task instructions
            task_message = self.task_instructions(task)
            self._add_message_with_tokens(task_message)

            # Add sensitive data information if provided
            if self.sensitive_data:
                info = f"Here are placeholders for sensitve data: {list(self.sensitive_data.keys())}"
                info += "To use them, write <secret>the placeholder name</secret>"
                info_message = HumanMessage(content=info)
                self._add_message_with_tokens(info_message)

            # Add example output placeholder
            placeholder_message = HumanMessage(content="Example output:")
            self._add_message_with_tokens(placeholder_message)

            # Add example tool call
            tool_calls = [
                {
                    'name': 'AgentOutput',
                    'args': {
                        'current_state': {
                            'page_summary': 'On the page are company a,b,c wtih their revenue 1,2,3.',
                            'evaluation_previous_goal': 'Success - I opend the first page',
                            'memory': 'Starting with the new task. I have completed 1/10 steps',
                            'next_goal': 'Click on company a'
                        },
                        'action': [
                            {
                                'click_element': {
                                    'index': 0
                                }
                            }
                        ]
                    },
                    'id': str(self.tool_id),
                    'type': 'tool_call'
                }
            ]
            example_ai_message = AIMessage(content='', tool_calls=tool_calls)
            self._add_message_with_tokens(example_ai_message)

            example_tool_message = ToolMessage(content='Browser started', tool_call_id=str(self.tool_id))
            self._add_message_with_tokens(example_tool_message)
            self.tool_id += 1

            memory_message = HumanMessage(content='[Your task history memory starts here]')
            self._add_message_with_tokens(memory_message)

    def _init_messages(self) -> None:
        """Initialize the message history with system message, context, task, and other initial messages"""
        self._add_message_with_tokens(self.system_prompt)

        if self.settings.message_context:
            context_message = HumanMessage(content='Context for the task' + self.settings.message_context)
            self._add_message_with_tokens(context_message)

        task_message = HumanMessage(
            content=f'Your ultimate task is: """{self.task}""". If you achieved your ultimate task, stop everything and use the done action in the next step to complete the task. If not, continue as usual.'
        )
        self._add_message_with_tokens(task_message)

        if self.settings.sensitive_data:
            info = f'Here are placeholders for sensitve data: {list(self.settings.sensitive_data.keys())}'
            info += 'To use them, write <secret>the placeholder name</secret>'
            info_message = HumanMessage(content=info)
            self._add_message_with_tokens(info_message)

        placeholder_message = HumanMessage(content='Example output:')
        self._add_message_with_tokens(placeholder_message)

        tool_calls = [
            {
                'name': 'AgentOutput',
                'args': {
                    'current_state': {
                        'evaluation_previous_goal': 'Success - I opend the first page',
                        'memory': 'Starting with the new task. I have completed 1/10 steps',
                        'next_goal': 'Click on company a',
                    },
                    'action': [{'click_element': {'index': 0}}],
                },
                'id': str(self.state.tool_id),
                'type': 'tool_call',
            }
        ]
        example_tool_call = AIMessage(content='', tool_calls=tool_calls)
        self._add_message_with_tokens(example_tool_call)
        self.add_tool_message(content='Browser started')

        placeholder_message = HumanMessage(content='[Your task history memory starts here]')
        self._add_message_with_tokens(placeholder_message)

        if self.settings.available_file_paths:
            filepaths_msg = HumanMessage(content=f'Here are file paths you can use: {self.settings.available_file_paths}')
            self._add_message_with_tokens(filepaths_msg)

    def task_instructions(self, task: str) -> HumanMessage:
        """Format task instructions as a human message."""
        content = f'''Your ultimate task is: """{task}""". If you achieved your ultimate task, stop everything and use the done action in the next step to complete the task. If not, continue as usual.'''
        return HumanMessage(content=content)

    def add_new_task(self, new_task: str) -> None:
        """Add a new task to the conversation."""
        content = f'''Your new ultimate task is: """{new_task}""". Take the previous context into account and finish your new ultimate task. '''
        msg = HumanMessage(content=content)
        self._add_message_with_tokens(msg)
        if hasattr(self, 'settings'):
            self.task = new_task

    def add_plan(self, plan: Optional[str], position: Optional[int] = None) -> None:
        """Add a plan as an AI message at the specified position."""
        if plan:
            msg = AIMessage(content=plan)
            self._add_message_with_tokens(msg, position)

    def add_state_message(
        self,
        state: BrowserState,
        result: Optional[List[ActionResult]] = None,
        step_info: Optional[AgentStepInfo] = None,
        use_vision: bool = True
    ) -> None:
        """Add browser state as human message"""
        if hasattr(self, 'settings'):
            if result:
                for r in result:
                    if r.include_in_memory:
                        if r.extracted_content:
                            msg = HumanMessage(content='Action result: ' + str(r.extracted_content))
                            self._add_message_with_tokens(msg)
                        if r.error:
                            if r.error.endswith('\n'):
                                r.error = r.error[:-1]
                            last_line = r.error.split('\n')[-1]
                            msg = HumanMessage(content='Action error: ' + last_line)
                            self._add_message_with_tokens(msg)
                        result = None
            state_message = AgentMessagePrompt(
                state,
                result,
                include_attributes=self.settings.include_attributes,
                step_info=step_info,
            ).get_user_message(use_vision)
            self._add_message_with_tokens(state_message)
        else:
            if result:
                for r in result:
                    if r.include_in_memory:
                        if r.extracted_content:
                            msg = HumanMessage(content='Action result: ' + str(r.extracted_content))
                            self._add_message_with_tokens(msg)
                        if r.error:
                            error = r.error[:self.max_error_length] if len(r.error) > self.max_error_length else r.error
                            if error.endswith('\n'):
                                error = error[:-1]
                            last_line = error.split('\n')[-1]
                            msg = HumanMessage(content='Action error: ' + last_line)
                            self._add_message_with_tokens(msg)
                        result = None
            state_message = AgentMessagePrompt(
                state,
                result,
                include_attributes=self.include_attributes,
                step_info=step_info,
            ).get_user_message(use_vision)
            self._add_message_with_tokens(state_message)

    def _remove_last_state_message(self) -> None:
        """Remove last state message from history"""
        if hasattr(self, 'state'):
            self.state.history.remove_last_state_message()
        else:
            if len(self.history.messages) > 2 and isinstance(self.history.messages[-1].message, HumanMessage):
                self.history.remove_message()

    def add_model_output(self, model_output: AgentOutput) -> None:
        """Add model output as AI message"""
        if hasattr(self, 'state'):
            tool_calls = [
                {
                    'name': 'AgentOutput',
                    'args': model_output.model_dump(mode='json', exclude_unset=True),
                    'id': str(self.state.tool_id),
                    'type': 'tool_call',
                }
            ]
            msg = AIMessage(content='', tool_calls=tool_calls)
            self._add_message_with_tokens(msg)
            self.add_tool_message(content='')
        else:
            tool_calls = [
                {
                    'name': 'AgentOutput',
                    'args': model_output.model_dump(mode='json', exclude_unset=True),
                    'id': str(self.tool_id),
                    'type': 'tool_call'
                }
            ]
            msg = AIMessage(content='', tool_calls=tool_calls)
            self._add_message_with_tokens(msg)
            tool_msg = ToolMessage(content='', tool_call_id=str(self.tool_id))
            self._add_message_with_tokens(tool_msg)
            self.tool_id += 1

    def add_tool_message(self, content: str) -> None:
        """Add tool message to history"""
        if hasattr(self, 'state'):
            msg = ToolMessage(content=content, tool_call_id=str(self.state.tool_id))
            self.state.tool_id += 1
            self._add_message_with_tokens(msg)
        else:
            msg = ToolMessage(content=content, tool_call_id=str(self.tool_id))
            self.tool_id += 1
            self._add_message_with_tokens(msg)

    def get_messages(self) -> List[BaseMessage]:
        """Get current message list, potentially trimmed to max tokens"""
        if hasattr(self, 'state'):
            messages = [m.message for m in self.state.history.messages]
            total_input_tokens = 0
            logger.debug(f'Messages in history: {len(self.state.history.messages)}:')
            for m in self.state.history.messages:
                total_input_tokens += m.metadata.tokens
                logger.debug(f'{m.message.__class__.__name__} - Token count: {m.metadata.tokens}')
            logger.debug(f'Total input tokens: {total_input_tokens}')
            return messages
        else:
            messages = [m.message for m in self.history.messages]
            total_tokens = 0
            logger.debug(f'Messages in history: {len(self.history.messages)}:')
            for m in self.history.messages:
                total_tokens += m.metadata.input_tokens
                logger.debug(f'{m.message.__class__.__name__} - Token count: {m.metadata.input_tokens}')
            logger.debug(f'Total input tokens: {total_tokens}')
            return messages

    def _add_message_with_tokens(self, message: BaseMessage, position: Optional[int] = None) -> None:
        """Add message with token count metadata
        position: None for last, -1 for second last, etc.
        """
        if hasattr(self, 'settings'):
            if self.settings.sensitive_data:
                message = self._filter_sensitive_data(message)
            token_count = self._count_tokens(message)
            metadata = MessageMetadata(tokens=token_count)
            self.state.history.add_message(message, metadata, position)
        else:
            if self.sensitive_data:
                message = self._filter_sensitive_data(message)
            token_count = self._count_tokens(message)
            metadata = MessageMetadata(input_tokens=token_count)
            self.history.add_message(message, metadata, position)

    def _filter_sensitive_data(self, message: BaseMessage) -> BaseMessage:
        """Filter out sensitive data from the message"""
        if hasattr(self, 'settings'):
            def replace_sensitive(value: str) -> str:
                if not self.settings.sensitive_data:
                    return value
                for key, val in self.settings.sensitive_data.items():
                    if not val:
                        continue
                    value = value.replace(val, f'<secret>{key}</secret>')
                return value

            if isinstance(message.content, str):
                message.content = replace_sensitive(message.content)
            elif isinstance(message.content, list):
                for i, item in enumerate(message.content):
                    if isinstance(item, dict) and 'text' in item:
                        item['text'] = replace_sensitive(item['text'])
                        message.content[i] = item
        else:
            def replace_sensitive(value: str) -> str:
                if not self.sensitive_data:
                    return value
                for key, val in self.sensitive_data.items():
                    if not val:
                        continue
                    value = value.replace(val, f'<secret>{key}</secret>')
                return value

            if isinstance(message.content, str):
                message.content = replace_sensitive(message.content)
            elif isinstance(message.content, list):
                for i, item in enumerate(message.content):
                    if isinstance(item, dict) and 'text' in item:
                        item['text'] = replace_sensitive(item['text'])
                        message.content[i] = item
        return message

    def _count_tokens(self, message: BaseMessage) -> int:
        """Count tokens in a message using the model's tokenizer"""
        tokens = 0
        if hasattr(self, 'settings'):
            if isinstance(message.content, list):
                for item in message.content:
                    if 'image_url' in item:
                        tokens += self.settings.image_tokens
                    elif isinstance(item, dict) and 'text' in item:
                        tokens += self._count_text_tokens(item['text'])
            else:
                msg = message.content
                if hasattr(message, 'tool_calls'):
                    msg += str(message.tool_calls)
                tokens += self._count_text_tokens(msg)
        else:
            if isinstance(message.content, list):
                for item in message.content:
                    if 'image_url' in item:
                        tokens += self.IMG_TOKENS
                    elif isinstance(item, dict) and 'text' in item:
                        tokens += self._count_text_tokens(item['text'])
            else:
                msg = message.content
                if hasattr(message, 'tool_calls'):
                    msg += str(message.tool_calls)
                tokens += self._count_text_tokens(msg)
        return tokens

    def _count_text_tokens(self, text: str) -> int:
        """Count tokens in a text string"""
        if hasattr(self, 'settings'):
            tokens = len(text) // self.settings.estimated_characters_per_token
        else:
            tokens = len(text) // self.estimated_characters_per_token
        return tokens

    def cut_messages(self):
        """Get current message list, potentially trimmed to max tokens"""
        if hasattr(self, 'state'):
            diff = self.state.history.current_tokens - self.settings.max_input_tokens
            if diff <= 0:
                return None

            msg = self.state.history.messages[-1]

            if isinstance(msg.message.content, list):
                text = ""
                for item in list(msg.message.content):
                    if 'image_url' in item:
                        msg.message.content.remove(item)
                        diff -= self.settings.image_tokens
                        msg.metadata.tokens -= self.settings.image_tokens
                        self.state.history.current_tokens -= self.settings.image_tokens
                        logger.debug(
                            f"Removed image with {self.settings.image_tokens} tokens - total tokens now: {self.state.history.current_tokens}/{self.settings.max_input_tokens}"
                        )
                    elif 'text' in item and isinstance(item, dict):
                        text += item['text']
                msg.message.content = text
                self.state.history.messages[-1] = msg

            if diff <= 0:
                return None

            proportion_to_remove = diff / msg.metadata.tokens
            if proportion_to_remove > 0.99:
                raise ValueError(
                    f"Max token limit reached - history is too long - reduce the system prompt or task. proportion_to_remove: {proportion_to_remove}"
                )
            logger.debug(
                f"Removing {proportion_to_remove * 100:.2f}% of the last message  {proportion_to_remove * msg.metadata.tokens:.2f} / {msg.metadata.tokens:.2f} tokens)"
            )

            content = msg.message.content
            characters_to_remove = int(len(content) * proportion_to_remove)
            content = content[:-characters_to_remove]

            self.state.history.remove_last_state_message()

            msg = HumanMessage(content=content)
            self._add_message_with_tokens(msg)

            last_msg = self.state.history.messages[-1]
            logger.debug(
                f"Added message with {last_msg.metadata.tokens} tokens - total tokens now: {self.state.history.current_tokens}/{self.settings.max_input_tokens} - total messages: {len(self.state.history.messages)}"
            )
        else:
            diff = self.history.total_tokens - self.max_input_tokens
            if diff <= 0:
                return None

            msg = self.history.messages[-1]

            if isinstance(msg.message.content, list):
                text = ""
                for item in list(msg.message.content):
                    if 'image_url' in item:
                        msg.message.content.remove(item)
                        diff -= self.IMG_TOKENS
                        msg.metadata.input_tokens -= self.IMG_TOKENS
                        self.history.total_tokens -= self.IMG_TOKENS
                        logger.debug(
                            f"Removed image with {self.IMG_TOKENS} tokens - total tokens now: {self.history.total_tokens}/{self.max_input_tokens}"
                        )
                    elif 'text' in item and isinstance(item, dict):
                        text += item['text']
                msg.message.content = text
                self.history.messages[-1] = msg

            if diff <= 0:
                return None

            proportion_to_remove = diff / msg.metadata.input_tokens
            if proportion_to_remove > 0.99:
                raise ValueError(
                    f"Max token limit reached - history is too long - reduce the system prompt or task. proportion_to_remove: {proportion_to_remove}"
                )
            logger.debug(
                f"Removing {proportion_to_remove * 100:.2f}% of the last message ({proportion_to_remove * msg.metadata.input_tokens:.2f} / {msg.metadata.input_tokens:.2f} tokens)"
            )

            content = msg.message.content
            characters_to_remove = int(len(content) * proportion_to_remove)
            content = content[:-characters_to_remove]

            self._remove_last_state_message()

            msg = HumanMessage(content=content)
            self._add_message_with_tokens(msg)

            last_msg = self.history.messages[-1]
            logger.debug(
                f"Added message with {last_msg.metadata.input_tokens} tokens - total tokens now: {self.history.total_tokens}/{self.max_input_tokens} - total messages: {len(self.history.messages)}"
            )

    def extract_json_from_model_output(self, content: str) -> dict:
        """Extract JSON from model output, handling both plain JSON and code-block-wrapped JSON."""
        if "```" in content:
            content = content.split("```")[1]
            if "\n" in content:
                content = content.split("\n", 1)[1]
        return json.loads(content)

    def convert_messages_for_non_function_calling_models(self, input_messages: List[BaseMessage]) -> List[BaseMessage]:
        """Convert messages for non-function-calling models"""
        converted_messages = []
        for msg in input_messages:
            if isinstance(msg, AIMessage) and hasattr(msg, 'tool_calls') and msg.tool_calls:
                tool_call = msg.tool_calls[0]
                content = json.dumps(tool_call.get('args', {}), indent=2)
                converted_messages.append(AIMessage(content=content))
            else:
                converted_messages.append(msg)
        return converted_messages

    def merge_successive_messages(self, messages: List[BaseMessage], class_to_merge: Type[BaseMessage]) -> List[BaseMessage]:
        """Some models like deepseek-reasoner dont allow multiple human messages in a row. This function merges them into one."""
        merged_messages = []
        buffer = ""
        buffer_items = 0

        for i, msg in enumerate(messages):
            if isinstance(msg, class_to_merge):
                if buffer_items > 0 and i < len(messages) - 1 and isinstance(messages[i+1], class_to_merge):
                    buffer += "\n\n" + msg.content
                    buffer_items += 1
                elif buffer_items > 0:
                    buffer += "\n\n" + msg.content
                    merged_messages.append(class_to_merge(content=buffer))
                    buffer = ""
                    buffer_items = 0
                else:
                    if i < len(messages) - 1 and isinstance(messages[i+1], class_to_merge):
                        buffer = msg.content
                        buffer_items = 1
                    else:
                        merged_messages.append(msg)
            else:
                if buffer_items > 0:
                    merged_messages.append(class_to_merge(content=buffer))
                    buffer = ""
                    buffer_items = 0
                merged_messages.append(msg)

        if buffer_items > 0:
            merged_messages.append(class_to_merge(content=buffer))
        return merged_messages



import logging
import os
import sys

from dotenv import load_dotenv

load_dotenv()


def addLoggingLevel(levelName, levelNum, methodName=None):
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example
    -------
    >>> addLoggingLevel('TRACE', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel('TRACE')
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.trace('so did this')
    >>> logging.TRACE
    5

    """
    if not methodName:
        methodName = levelName.lower()

    if hasattr(logging, levelName):
        raise AttributeError('{} already defined in logging module'.format(levelName))
    if hasattr(logging, methodName):
        raise AttributeError('{} already defined in logging module'.format(methodName))
    if hasattr(logging.getLoggerClass(), methodName):
        raise AttributeError('{} already defined in logger class'.format(methodName))

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def logForLevel(self, message, *args, **kwargs):
        if self.isEnabledFor(levelNum):
            self._log(levelNum, message, args, **kwargs)

    def logToRoot(message, *args, **kwargs):
        logging.log(levelNum, message, *args, **kwargs)

    logging.addLevelName(levelNum, levelName)
    setattr(logging, levelName, levelNum)
    setattr(logging.getLoggerClass(), methodName, logForLevel)
    setattr(logging, methodName, logToRoot)


def setup_logging():
    # Try to add RESULT level, but ignore if it already exists
    try:
        addLoggingLevel('RESULT', 35)  # This allows ERROR, FATAL and CRITICAL
    except AttributeError:
        pass  # Level already exists, which is fine

    log_type = os.getenv('BROWSER_USE_LOGGING_LEVEL', 'info').lower()

    # Check if handlers are already set up
    if logging.getLogger().hasHandlers():
        return

    # Clear existing handlers
    root = logging.getLogger()
    root.handlers = []

    class BrowserUseFormatter(logging.Formatter):
        def format(self, record):
            if type(record.name) == str and record.name.startswith('browser_use.'):
                record.name = record.name.split('.')[-2]
            return super().format(record)

    # Setup single handler for all loggers
    console = logging.StreamHandler(sys.stdout)

    # additional setLevel here to filter logs
    if log_type == 'result':
        console.setLevel('RESULT')
        console.setFormatter(BrowserUseFormatter('%(message)s'))
    else:
        console.setFormatter(BrowserUseFormatter('%(levelname)-8s [%(name)s] %(message)s'))

    # Configure root logger only
    root.addHandler(console)

    # switch cases for log_type
    if log_type == 'result':
        root.setLevel('RESULT')  # string usage to avoid syntax error
    elif log_type == 'debug':
        root.setLevel(logging.DEBUG)
    else:
        root.setLevel(logging.INFO)

    # Configure browser_use logger
    browser_use_logger = logging.getLogger('browser_use')
    browser_use_logger.propagate = False  # Don't propagate to root logger
    browser_use_logger.addHandler(console)
    browser_use_logger.setLevel(root.level)  # Set same level as root logger

    logger = logging.getLogger('browser_use')
    logger.info('BrowserUse logging setup complete with level %s', log_type)
    
    # Silence third-party loggers
    for logger in [
        'WDM',
        'httpx',
        'selenium',
        'playwright',
        'urllib3',
        'asyncio',
        'langchain',
        'openai',
        'httpcore',
        'charset_normalizer',
        'anthropic._base_client',
        'PIL.PngImagePlugin',
        'trafilatura.htmlprocessing',
        'trafilatura',
    ]:
        third_party = logging.getLogger(logger)
        third_party.setLevel(logging.ERROR)
        third_party.propagate = False


from typing import Literal
from pydantic import BaseModel
from app.types.browser_types import BrowserAction, BrowserActionResult

TextEditorCommand = Literal['dir', 'view', 'create', 'write', 'replace', 'find_content', 'find_file', 'move', 'delete']

class CommonApiResult(BaseModel):
    status: Literal['success', 'error']
    error: str | None = None

class FileActionResult(BaseModel):
    status: Literal['success', 'error']
    result: str

class TextEditorDirRequest(BaseModel):
    path: str
    sudo: bool | None = None

class TextEditorViewRequest(BaseModel):
    path: str
    sudo: bool | None = None
    view_range: list[int] | None = None

class TextEditorCreateRequest(BaseModel):
    path: str
    sudo: bool | None = None
    file_text: str

class TextEditorWriteRequest(BaseModel):
    path: str
    sudo: bool | None = None
    file_text: str
    append: bool | None = None
    leading_newline: bool | None = None
    trailing_newline: bool | None = None

class TextEditorStrReplaceRequest(BaseModel):
    path: str
    sudo: bool | None = None
    old_str: str
    new_str: str

class TextEditorFindContentRequest(BaseModel):
    path: str
    sudo: bool | None = None
    regex: str

class TextEditorFindFileRequest(BaseModel):
    path: str
    sudo: bool | None = None
    glob: str

class TextEditorMoveRequest(BaseModel):
    path: str
    sudo: bool | None = None
    new_path: str

class TextEditorDeleteRequest(BaseModel):
    path: str
    sudo: bool | None = None

class FileInfo(BaseModel):
    path: str
    content: str
    old_content: str | None = None

class TextEditorActionResult(CommonApiResult):
    result: str
    file_info: FileInfo | None = None

class BrowserActionRequest(BaseModel):
    action: BrowserAction
    screenshot_presigned_url: str | None = None
    clean_screenshot_presigned_url: str | None = None

class BrowserActionResponse(CommonApiResult):
    result: BrowserActionResult | None = None

class TerminalWriteApiRequest(BaseModel):
    text: str
    enter: bool | None = None

class TerminalApiResponse(CommonApiResult):
    output: list[str]
    result: str
    terminal_id: str

TerminalInputMessageType = Literal['command', 'view', 'view_last', 'kill', 'reset', 'reset_all']
TerminalOutputMessageType = Literal['update', 'finish', 'partial_finish', 'error', 'history', 'action_finish']
TerminalCommandMode = Literal['run', 'send_line', 'send_key', 'send_control']
TerminalStatus = Literal['idle', 'running']

class TerminalInputMessage(BaseModel):
    type: TerminalInputMessageType
    terminal: str
    action_id: str
    command: str | None = None
    mode: TerminalCommandMode | None = None
    exec_dir: str | None = None

    def create_response(self, type: TerminalOutputMessageType, result: str, output: list[str], terminal_status: TerminalStatus, sub_command_index: int = 0):
        return TerminalOutputMessage(
            type=type,
            terminal=self.terminal,
            action_id=self.action_id,
            sub_command_index=sub_command_index,
            result=result,
            output=output,
            terminal_status=terminal_status
        )

class TerminalOutputMessage(BaseModel):
    type: TerminalOutputMessageType
    terminal: str
    action_id: str
    sub_command_index: int = 0 
    result: str | None = None
    output: list[str]
    terminal_status: Literal['idle', 'running', 'unknown']
